# The MIT License
#
# Copyright (c) 2015 the bpython authors.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pydoc
from types import TracebackType
from typing import Literal

from .. import _internal


class NopPydocPager:
    def __enter__(self):
        self._orig_pager = pydoc.pager
        pydoc.pager = self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        pydoc.pager = self._orig_pager
        return False

    def __call__(self, text):
        return None


class _Helper(_internal._Helper):
    def __init__(self, repl=None):
        self._repl = repl
        pydoc.pager = self.pager

        super().__init__()

    def pager(self, output, title=""):
        self._repl.pager(output, title)

    def __call__(self, *args, **kwargs):
        if self._repl.reevaluating:
            with NopPydocPager():
                return super().__call__(*args, **kwargs)
        else:
            return super().__call__(*args, **kwargs)


# vim: sw=4 ts=4 sts=4 ai et



# The MIT License
#
# Copyright (c) 2009-2011 Andreas Stuehrk
# Copyright (c) 2020-2021 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import fnmatch
import importlib.machinery
import sys
import warnings
from dataclasses import dataclass
from pathlib import Path
from collections.abc import Generator, Sequence, Iterable

from .line import (
    current_word,
    current_import,
    current_from_import_from,
    current_from_import_import,
)

SUFFIXES = importlib.machinery.all_suffixes()
LOADERS = (
    (
        importlib.machinery.ExtensionFileLoader,
        importlib.machinery.EXTENSION_SUFFIXES,
    ),
    (
        importlib.machinery.SourceFileLoader,
        importlib.machinery.SOURCE_SUFFIXES,
    ),
)

_LOADED_INODE_DATACLASS_ARGS = {"frozen": True}
if sys.version_info[:2] >= (3, 10):
    _LOADED_INODE_DATACLASS_ARGS["slots"] = True


@dataclass(**_LOADED_INODE_DATACLASS_ARGS)
class _LoadedInode:
    dev: int
    inode: int


class ModuleGatherer:
    def __init__(
        self,
        paths: Iterable[str | Path] | None = None,
        skiplist: Sequence[str] | None = None,
    ) -> None:
        """Initialize module gatherer with all modules in `paths`, which should be a list of
        directory names. If `paths` is not given, `sys.path` will be used."""

        # Cached list of all known modules
        self.modules: set[str] = set()
        # Set of (st_dev, st_ino) to compare against so that paths are not repeated
        self.paths: set[_LoadedInode] = set()
        # Patterns to skip
        self.skiplist: Sequence[str] = (
            skiplist if skiplist is not None else tuple()
        )
        self.fully_loaded = False

        if paths is None:
            self.modules.update(sys.builtin_module_names)
            paths = sys.path

        self.find_iterator = self.find_all_modules(
            Path(p).resolve() if p else Path.cwd() for p in paths
        )

    def module_matches(self, cw: str, prefix: str = "") -> set[str]:
        """Modules names to replace cw with"""

        full = f"{prefix}.{cw}" if prefix else cw
        matches = (
            name
            for name in self.modules
            if (name.startswith(full) and name.find(".", len(full)) == -1)
        )
        if prefix:
            return {match[len(prefix) + 1 :] for match in matches}
        else:
            return set(matches)

    def attr_matches(
        self, cw: str, prefix: str = "", only_modules: bool = False
    ) -> set[str]:
        """Attributes to replace name with"""
        full = f"{prefix}.{cw}" if prefix else cw
        module_name, _, name_after_dot = full.rpartition(".")
        if module_name not in sys.modules:
            return set()
        module = sys.modules[module_name]
        if only_modules:
            matches = {
                name
                for name in dir(module)
                if name.startswith(name_after_dot)
                and f"{module_name}.{name}" in sys.modules
            }
        else:
            matches = {
                name for name in dir(module) if name.startswith(name_after_dot)
            }
        module_part = cw.rpartition(".")[0]
        if module_part:
            matches = {f"{module_part}.{m}" for m in matches}

        return matches

    def module_attr_matches(self, name: str) -> set[str]:
        """Only attributes which are modules to replace name with"""
        return self.attr_matches(name, only_modules=True)

    def complete(self, cursor_offset: int, line: str) -> set[str] | None:
        """Construct a full list of possibly completions for imports."""
        tokens = line.split()
        if "from" not in tokens and "import" not in tokens:
            return None

        result = current_word(cursor_offset, line)
        if result is None:
            return None

        from_import_from = current_from_import_from(cursor_offset, line)
        if from_import_from is not None:
            import_import = current_from_import_import(cursor_offset, line)
            if import_import is not None:
                # `from a import <b|>` completion
                matches = self.module_matches(
                    import_import.word, from_import_from.word
                )
                matches.update(
                    self.attr_matches(import_import.word, from_import_from.word)
                )
            else:
                # `from <a|>` completion
                matches = self.module_attr_matches(from_import_from.word)
                matches.update(self.module_matches(from_import_from.word))
            return matches

        cur_import = current_import(cursor_offset, line)
        if cur_import is not None:
            # `import <a|>` completion
            matches = self.module_matches(cur_import.word)
            matches.update(self.module_attr_matches(cur_import.word))
            return matches
        else:
            return None

    def find_modules(self, path: Path) -> Generator[str | None, None, None]:
        """Find all modules (and packages) for a given directory."""
        if not path.is_dir():
            # Perhaps a zip file
            return
        if any(fnmatch.fnmatch(path.name, entry) for entry in self.skiplist):
            # Path is on skiplist
            return

        finder = importlib.machinery.FileFinder(str(path), *LOADERS)  # type: ignore
        try:
            for p in path.iterdir():
                if p.name.startswith(".") or p.name == "__pycache__":
                    # Impossible to import from names starting with . and we can skip __pycache__
                    continue
                elif any(
                    fnmatch.fnmatch(p.name, entry) for entry in self.skiplist
                ):
                    # Path is on skiplist
                    continue
                elif not any(p.name.endswith(suffix) for suffix in SUFFIXES):
                    # Possibly a package
                    if "." in p.name:
                        continue
                elif p.is_dir():
                    # Unfortunately, CPython just crashes if there is a directory
                    # which ends with a python extension, so work around.
                    continue
                name = p.name
                for suffix in SUFFIXES:
                    if name.endswith(suffix):
                        name = name[: -len(suffix)]
                        break
                if name == "badsyntax_pep3120":
                    # Workaround for issue #166
                    continue

                package_pathname = None
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", ImportWarning)
                        spec = finder.find_spec(name)
                        if spec is None:
                            continue
                        if spec.submodule_search_locations is not None:
                            package_pathname = spec.submodule_search_locations[
                                0
                            ]
                except (ImportError, OSError, SyntaxError, UnicodeEncodeError):
                    # UnicodeEncodeError happens with Python 3 when there is a filename in some invalid encoding
                    continue

                if package_pathname is not None:
                    path_real = Path(package_pathname).resolve()
                    try:
                        stat = path_real.stat()
                    except OSError:
                        continue
                    loaded_inode = _LoadedInode(stat.st_dev, stat.st_ino)
                    if loaded_inode not in self.paths:
                        self.paths.add(loaded_inode)
                        for subname in self.find_modules(path_real):
                            if subname is None:
                                yield None  # take a break to avoid unresponsiveness
                            elif subname != "__init__":
                                yield f"{name}.{subname}"
                yield name
        except OSError:
            # Path is not readable
            return
        yield None  # take a break to avoid unresponsiveness

    def find_all_modules(
        self, paths: Iterable[Path]
    ) -> Generator[None, None, None]:
        """Return a list with all modules in `path`, which should be a list of
        directory names. If path is not given, sys.path will be used."""

        for p in paths:
            for module in self.find_modules(p):
                if module is not None:
                    self.modules.add(module)
                yield

    def find_coroutine(self) -> bool:
        if self.fully_loaded:
            return False

        try:
            next(self.find_iterator)
        except StopIteration:
            self.fully_loaded = True

        return True



import re
from functools import partial
from typing import Any
from collections.abc import Callable

from curtsies.formatstring import fmtstr, FmtStr
from curtsies.termformatconstants import (
    FG_COLORS,
    BG_COLORS,
    colors as CURTSIES_COLORS,
)

from ..config import COLOR_LETTERS
from ..lazyre import LazyReCompile


COLORS = CURTSIES_COLORS + ("default",)
CNAMES = dict(zip(COLOR_LETTERS, COLORS))
# hack for finding the "inverse"
INVERSE_COLORS = {
    CURTSIES_COLORS[idx]: CURTSIES_COLORS[
        (idx + (len(CURTSIES_COLORS) // 2)) % len(CURTSIES_COLORS)
    ]
    for idx in range(len(CURTSIES_COLORS))
}
INVERSE_COLORS["default"] = INVERSE_COLORS[CURTSIES_COLORS[0]]


def func_for_letter(
    letter_color_code: str, default: str = "k"
) -> Callable[..., FmtStr]:
    """Returns FmtStr constructor for a bpython-style color code"""
    if letter_color_code == "d":
        letter_color_code = default
    elif letter_color_code == "D":
        letter_color_code = default.upper()
    return partial(
        fmtstr,
        fg=CNAMES[letter_color_code.lower()],
        bold=letter_color_code.isupper(),
    )


def color_for_letter(letter_color_code: str, default: str = "k") -> str:
    if letter_color_code == "d":
        letter_color_code = default
    return CNAMES[letter_color_code.lower()]


def parse(s: str) -> FmtStr:
    """Returns a FmtStr object from a bpython-formatted colored string"""
    rest = s
    stuff = []
    while rest:
        start, rest = peel_off_string(rest)
        stuff.append(start)
    return (
        sum((fs_from_match(d) for d in stuff[1:]), fs_from_match(stuff[0]))
        if len(stuff) > 0
        else FmtStr()
    )


def fs_from_match(d: dict[str, Any]) -> FmtStr:
    atts = {}
    color = "default"
    if d["fg"]:
        # this isn't according to spec as I understand it
        if d["fg"].isupper():
            d["bold"] = True
        # TODO figure out why boldness isn't based on presence of \x02

        color = CNAMES[d["fg"].lower()]
        if color != "default":
            atts["fg"] = FG_COLORS[color]
    if d["bg"]:
        if d["bg"] == "I":
            # hack for finding the "inverse"
            color = INVERSE_COLORS[color]
        else:
            color = CNAMES[d["bg"].lower()]
        if color != "default":
            atts["bg"] = BG_COLORS[color]
    if d["bold"]:
        atts["bold"] = True
    return fmtstr(d["string"], **atts)


peel_off_string_re = LazyReCompile(
    r"""(?P<colormarker>\x01
            (?P<fg>[krgybmcwdKRGYBMCWD]?)
            (?P<bg>[krgybmcwdKRGYBMCWDI]?)?)
        (?P<bold>\x02?)
        \x03
        (?P<string>[^\x04]*)
        \x04
        (?P<rest>.*)
        """,
    re.VERBOSE | re.DOTALL,
)


def peel_off_string(s: str) -> tuple[dict[str, Any], str]:
    m = peel_off_string_re.match(s)
    assert m, repr(s)
    d = m.groupdict()
    rest = d["rest"]
    del d["rest"]
    return d, rest



import inspect
import keyword
import unittest
from collections import namedtuple
from unittest import mock

try:
    import jedi

    has_jedi = True
except ImportError:
    has_jedi = False

from bpython import autocomplete, inspection
from bpython.line import LinePart

glob_function = "glob.iglob"


class TestSafeEval(unittest.TestCase):
    def test_catches_syntax_error(self):
        with self.assertRaises(autocomplete.EvaluationError):
            autocomplete.safe_eval("1re", {})


class TestFormatters(unittest.TestCase):
    def test_filename(self):
        completer = autocomplete.FilenameCompletion()
        last_part_of_filename = completer.format
        self.assertEqual(last_part_of_filename("abc"), "abc")
        self.assertEqual(last_part_of_filename("abc/"), "abc/")
        self.assertEqual(last_part_of_filename("abc/efg"), "efg")
        self.assertEqual(last_part_of_filename("abc/efg/"), "efg/")
        self.assertEqual(last_part_of_filename("/abc"), "abc")
        self.assertEqual(last_part_of_filename("ab.c/e.f.g/"), "e.f.g/")

    def test_attribute(self):
        self.assertEqual(autocomplete._after_last_dot("abc.edf"), "edf")


def completer(matches):
    mock_completer = autocomplete.BaseCompletionType()
    mock_completer.matches = mock.Mock(return_value=matches)
    return mock_completer


class TestGetCompleter(unittest.TestCase):
    def test_no_completers(self):
        self.assertTupleEqual(autocomplete.get_completer([], 0, ""), ([], None))

    def test_one_completer_without_matches_returns_empty_list_and_none(self):
        a = completer([])
        self.assertTupleEqual(
            autocomplete.get_completer([a], 0, ""), ([], None)
        )

    def test_one_completer_returns_matches_and_completer(self):
        a = completer(["a"])
        self.assertTupleEqual(
            autocomplete.get_completer([a], 0, ""), (["a"], a)
        )

    def test_two_completers_with_matches_returns_first_matches(self):
        a = completer(["a"])
        b = completer(["b"])
        self.assertEqual(autocomplete.get_completer([a, b], 0, ""), (["a"], a))

    def test_first_non_none_completer_matches_are_returned(self):
        a = completer([])
        b = completer(["a"])
        self.assertEqual(autocomplete.get_completer([a, b], 0, ""), ([], None))

    def test_only_completer_returns_None(self):
        a = completer(None)
        self.assertEqual(autocomplete.get_completer([a], 0, ""), ([], None))

    def test_first_completer_returns_None(self):
        a = completer(None)
        b = completer(["a"])
        self.assertEqual(autocomplete.get_completer([a, b], 0, ""), (["a"], b))


class TestCumulativeCompleter(unittest.TestCase):
    def completer(self, matches):
        mock_completer = autocomplete.BaseCompletionType()
        mock_completer.matches = mock.Mock(return_value=matches)
        return mock_completer

    def test_no_completers_fails(self):
        with self.assertRaises(ValueError):
            autocomplete.CumulativeCompleter([])

    def test_one_empty_completer_returns_empty(self):
        a = self.completer([])
        cumulative = autocomplete.CumulativeCompleter([a])
        self.assertEqual(cumulative.matches(3, "abc"), set())

    def test_one_none_completer_returns_none(self):
        a = self.completer(None)
        cumulative = autocomplete.CumulativeCompleter([a])
        self.assertEqual(cumulative.matches(3, "abc"), None)

    def test_two_completers_get_both(self):
        a = self.completer(["a"])
        b = self.completer(["b"])
        cumulative = autocomplete.CumulativeCompleter([a, b])
        self.assertEqual(cumulative.matches(3, "abc"), {"a", "b"})

    def test_order_completer(self):
        a = self.completer(["ax", "ab="])
        b = self.completer(["aa"])
        cumulative = autocomplete.CumulativeCompleter([a, b])
        self.assertEqual(
            autocomplete.get_completer([cumulative], 1, "a"),
            (["ab=", "aa", "ax"], cumulative),
        )


class TestFilenameCompletion(unittest.TestCase):
    def setUp(self):
        self.completer = autocomplete.FilenameCompletion()

    def test_locate_fails_when_not_in_string(self):
        self.assertEqual(self.completer.locate(4, "abcd"), None)

    def test_locate_succeeds_when_in_string(self):
        self.assertEqual(
            self.completer.locate(4, "a'bc'd"), LinePart(2, 4, "bc")
        )

    def test_issue_491(self):
        self.assertNotEqual(self.completer.matches(9, '"a[a.l-1]'), None)

    @mock.patch(glob_function, new=lambda text: [])
    def test_match_returns_none_if_not_in_string(self):
        self.assertEqual(self.completer.matches(2, "abcd"), None)

    @mock.patch(glob_function, new=lambda text: [])
    def test_match_returns_empty_list_when_no_files(self):
        self.assertEqual(self.completer.matches(2, '"a'), set())

    @mock.patch(glob_function, new=lambda text: ["abcde", "aaaaa"])
    @mock.patch("os.path.expanduser", new=lambda text: text)
    @mock.patch("os.path.isdir", new=lambda text: False)
    @mock.patch("os.path.sep", new="/")
    def test_match_returns_files_when_files_exist(self):
        self.assertEqual(
            sorted(self.completer.matches(2, '"x')), ["aaaaa", "abcde"]
        )

    @mock.patch(glob_function, new=lambda text: ["abcde", "aaaaa"])
    @mock.patch("os.path.expanduser", new=lambda text: text)
    @mock.patch("os.path.isdir", new=lambda text: True)
    @mock.patch("os.path.sep", new="/")
    def test_match_returns_dirs_when_dirs_exist(self):
        self.assertEqual(
            sorted(self.completer.matches(2, '"x')), ["aaaaa/", "abcde/"]
        )

    @mock.patch(
        glob_function, new=lambda text: ["/expand/ed/abcde", "/expand/ed/aaaaa"]
    )
    @mock.patch(
        "os.path.expanduser", new=lambda text: text.replace("~", "/expand/ed")
    )
    @mock.patch("os.path.isdir", new=lambda text: False)
    @mock.patch("os.path.sep", new="/")
    def test_tilde_stays_pretty(self):
        self.assertEqual(
            sorted(self.completer.matches(4, '"~/a')), ["~/aaaaa", "~/abcde"]
        )

    @mock.patch("os.path.sep", new="/")
    def test_formatting_takes_just_last_part(self):
        self.assertEqual(self.completer.format("/hello/there/"), "there/")
        self.assertEqual(self.completer.format("/hello/there"), "there")


class MockNumPy:
    """This is a mock numpy object that raises an error when there is an attempt
    to convert it to a boolean."""

    def __nonzero__(self):
        raise ValueError(
            "The truth value of an array with more than one "
            "element is ambiguous. Use a.any() or a.all()"
        )


class TestDictKeyCompletion(unittest.TestCase):
    def test_set_of_keys_returned_when_matches_found(self):
        com = autocomplete.DictKeyCompletion()
        local = {"d": {"ab": 1, "cd": 2}}
        self.assertSetEqual(
            com.matches(2, "d[", locals_=local), {"'ab']", "'cd']"}
        )

    def test_none_returned_when_eval_error(self):
        com = autocomplete.DictKeyCompletion()
        local = {"e": {"ab": 1, "cd": 2}}
        self.assertEqual(com.matches(2, "d[", locals_=local), None)

    def test_none_returned_when_not_dict_type(self):
        com = autocomplete.DictKeyCompletion()
        local = {"l": ["ab", "cd"]}
        self.assertEqual(com.matches(2, "l[", locals_=local), None)

    def test_none_returned_when_no_matches_left(self):
        com = autocomplete.DictKeyCompletion()
        local = {"d": {"ab": 1, "cd": 2}}
        self.assertEqual(com.matches(3, "d[r", locals_=local), None)

    def test_obj_that_does_not_allow_conversion_to_bool(self):
        com = autocomplete.DictKeyCompletion()
        local = {"mNumPy": MockNumPy()}
        self.assertEqual(com.matches(7, "mNumPy[", locals_=local), None)


class Foo:
    a = 10

    def __init__(self):
        self.b = 20

    def method(self, x):
        pass


class Properties(Foo):
    @property
    def asserts_when_called(self):
        raise AssertionError("getter method called")


class Slots:
    __slots__ = ["a", "b"]


class OverriddenGetattribute(Foo):
    def __getattribute__(self, name):
        raise AssertionError("custom get attribute invoked")


class TestAttrCompletion(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.com = autocomplete.AttrCompletion()

    def test_att_matches_found_on_instance(self):
        self.assertSetEqual(
            self.com.matches(2, "a.", locals_={"a": Foo()}),
            {"a.method", "a.a", "a.b"},
        )

    def test_descriptor_attributes_not_run(self):
        com = autocomplete.AttrCompletion()
        self.assertSetEqual(
            com.matches(2, "a.", locals_={"a": Properties()}),
            {"a.b", "a.a", "a.method", "a.asserts_when_called"},
        )

    def test_custom_get_attribute_not_invoked(self):
        com = autocomplete.AttrCompletion()
        self.assertSetEqual(
            com.matches(2, "a.", locals_={"a": OverriddenGetattribute()}),
            {"a.b", "a.a", "a.method"},
        )

    def test_slots_not_crash(self):
        com = autocomplete.AttrCompletion()
        self.assertSetEqual(
            com.matches(2, "A.", locals_={"A": Slots}),
            {"A.b", "A.a"},
        )


class TestExpressionAttributeCompletion(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.com = autocomplete.ExpressionAttributeCompletion()

    def test_att_matches_found_on_instance(self):
        self.assertSetEqual(
            self.com.matches(5, "a[0].", locals_={"a": [Foo()]}),
            {"method", "a", "b"},
        )

    def test_other_getitem_methods_not_called(self):
        class FakeList:
            def __getitem__(inner_self, i):
                self.fail("possibly side-effecting __getitem_ method called")

        self.com.matches(5, "a[0].", locals_={"a": FakeList()})

    def test_tuples_complete(self):
        self.assertSetEqual(
            self.com.matches(5, "a[0].", locals_={"a": (Foo(),)}),
            {"method", "a", "b"},
        )

    @unittest.skip("TODO, subclasses do not complete yet")
    def test_list_subclasses_complete(self):
        class ListSubclass(list):
            pass

        self.assertSetEqual(
            self.com.matches(5, "a[0].", locals_={"a": ListSubclass([Foo()])}),
            {"method", "a", "b"},
        )

    def test_getitem_not_called_in_list_subclasses_overriding_getitem(self):
        class FakeList(list):
            def __getitem__(inner_self, i):
                self.fail("possibly side-effecting __getitem_ method called")

        self.com.matches(5, "a[0].", locals_={"a": FakeList()})

    def test_literals_complete(self):
        self.assertSetEqual(
            self.com.matches(10, "[a][0][0].", locals_={"a": (Foo(),)}),
            {"method", "a", "b"},
        )

    def test_dictionaries_complete(self):
        self.assertSetEqual(
            self.com.matches(7, 'a["b"].', locals_={"a": {"b": Foo()}}),
            {"method", "a", "b"},
        )


class TestMagicMethodCompletion(unittest.TestCase):
    def test_magic_methods_complete_after_double_underscores(self):
        com = autocomplete.MagicMethodCompletion()
        block = "class Something(object)\n    def __"
        self.assertSetEqual(
            com.matches(
                10,
                "    def __",
                current_block=block,
                complete_magic_methods=True,
            ),
            set(autocomplete.MAGIC_METHODS),
        )


Completion = namedtuple("Completion", ["name", "complete"])


@unittest.skipUnless(has_jedi, "jedi required")
class TestMultilineJediCompletion(unittest.TestCase):
    def test_returns_none_with_single_line(self):
        com = autocomplete.MultilineJediCompletion()
        self.assertEqual(
            com.matches(2, "Va", current_block="Va", history=[]), None
        )

    def test_returns_non_with_blank_second_line(self):
        com = autocomplete.MultilineJediCompletion()
        self.assertEqual(
            com.matches(
                0, "", current_block="class Foo():\n", history=["class Foo():"]
            ),
            None,
        )

    def matches_from_completions(
        self, cursor, line, block, history, completions
    ):
        with mock.patch("bpython.autocomplete.jedi.Script") as Script:
            script = Script.return_value
            script.complete.return_value = completions
            com = autocomplete.MultilineJediCompletion()
            return com.matches(
                cursor, line, current_block=block, history=history
            )

    def test_completions_starting_with_different_letters(self):
        matches = self.matches_from_completions(
            2,
            " a",
            "class Foo:\n a",
            ["adsf"],
            [Completion("Abc", "bc"), Completion("Cbc", "bc")],
        )
        self.assertEqual(matches, None)

    def test_completions_starting_with_different_cases(self):
        matches = self.matches_from_completions(
            2,
            " a",
            "class Foo:\n a",
            ["adsf"],
            [Completion("Abc", "bc"), Completion("ade", "de")],
        )
        self.assertSetEqual(matches, {"ade"})

    def test_issue_544(self):
        com = autocomplete.MultilineJediCompletion()
        code = "@asyncio.coroutine\ndef"
        history = ("import asyncio", "@asyncio.coroutin")
        com.matches(3, "def", current_block=code, history=history)


class TestGlobalCompletion(unittest.TestCase):
    def setUp(self):
        self.com = autocomplete.GlobalCompletion()

    def test_function(self):
        def function():
            pass

        self.assertEqual(
            self.com.matches(8, "function", locals_={"function": function}),
            {"function("},
        )

    def test_completions_are_unicode(self):
        for m in self.com.matches(1, "a", locals_={"abc": 10}):
            self.assertIsInstance(m, str)

    def test_mock_kwlist(self):
        with mock.patch.object(keyword, "kwlist", new=["abcd"]):
            self.assertEqual(self.com.matches(3, "abc", locals_={}), None)

    def test_mock_kwlist_non_ascii(self):
        with mock.patch.object(keyword, "kwlist", new=["abcß"]):
            self.assertEqual(self.com.matches(3, "abc", locals_={}), None)


class TestParameterNameCompletion(unittest.TestCase):
    def test_set_of_params_returns_when_matches_found(self):
        def func(apple, apricot, banana, carrot):
            pass

        argspec = inspection.ArgSpec(*inspect.getfullargspec(func))
        funcspec = inspection.FuncProps("func", argspec, False)
        com = autocomplete.ParameterNameCompletion()
        self.assertSetEqual(
            com.matches(1, "a", funcprops=funcspec), {"apple=", "apricot="}
        )
        self.assertSetEqual(
            com.matches(2, "ba", funcprops=funcspec), {"banana="}
        )
        self.assertSetEqual(
            com.matches(3, "car", funcprops=funcspec), {"carrot="}
        )
        self.assertSetEqual(
            com.matches(5, "func(", funcprops=funcspec),
            {"apple=", "apricot=", "banana=", "carrot="},
        )



"""Extracting and changing portions of the current line

All functions take cursor offset from the beginning of the line and the line of
Python code, and return None, or a tuple of the start index, end index, and the
word."""

import re

from dataclasses import dataclass
from itertools import chain

from .lazyre import LazyReCompile


@dataclass
class LinePart:
    start: int
    stop: int
    word: str


_current_word_re = LazyReCompile(r"(?<![)\]\w_.])" r"([\w_][\w0-9._]*[(]?)")
CHARACTER_PAIR_MAP = {"(": ")", "{": "}", "[": "]", "'": "'", '"': '"'}


def current_word(cursor_offset: int, line: str) -> LinePart | None:
    """the object.attribute.attribute just before or under the cursor"""
    start = cursor_offset
    end = cursor_offset
    word = None
    for m in _current_word_re.finditer(line):
        if m.start(1) < cursor_offset <= m.end(1):
            start = m.start(1)
            end = m.end(1)
            word = m.group(1)
    if word is None:
        return None
    return LinePart(start, end, word)


# pieces of regex to match repr() of several hashable built-in types
_match_all_dict_keys = r"""[^\]]*"""

# https://docs.python.org/3/reference/lexical_analysis.html#string-and-bytes-literals
_match_single_quote_str_bytes = r"""
    # bytes repr() begins with `b` character; bytes and str begin with `'`
    b?'             
    # match escape sequence; this handles `\'` in the string repr()
    (?:\\['"nabfrtvxuU\\]|
    # or match any non-`\` and non-single-quote character (most of the string)
    [^'\\])*
    # matches hanging `\` or ending `'` if one is present
    [\\']?
"""

# bytes and str repr() only uses double quotes if the string contains 1 or more
# `'` character and exactly 0 `"` characters
_match_double_quote_str_bytes = r"""
    # bytes repr() begins with `b` character
    b?"
    # string continues until a `"` character is reached
    [^"]*
    # end matching at closing double-quote if one is present
    "?"""

# match valid identifier name followed by `[` character
_match_dict_before_key = r"""[\w_][\w0-9._]*\["""

_current_dict_key_re = LazyReCompile(
    f"{_match_dict_before_key}((?:"
    f"{_match_single_quote_str_bytes}|"
    f"{_match_double_quote_str_bytes}|"
    f"{_match_all_dict_keys}|)*)",
    re.VERBOSE,
)


def current_dict_key(cursor_offset: int, line: str) -> LinePart | None:
    """If in dictionary completion, return the current key"""
    for m in _current_dict_key_re.finditer(line):
        if m.start(1) <= cursor_offset <= m.end(1):
            return LinePart(m.start(1), m.end(1), m.group(1))
    return None


# capture valid identifier name if followed by `[` character
_capture_dict_name = r"""([\w_][\w0-9._]*)\["""

_current_dict_re = LazyReCompile(
    f"{_capture_dict_name}((?:"
    f"{_match_single_quote_str_bytes}|"
    f"{_match_double_quote_str_bytes}|"
    f"{_match_all_dict_keys}|)*)",
    re.VERBOSE,
)


def current_dict(cursor_offset: int, line: str) -> LinePart | None:
    """If in dictionary completion, return the dict that should be used"""
    for m in _current_dict_re.finditer(line):
        if m.start(2) <= cursor_offset <= m.end(2):
            return LinePart(m.start(1), m.end(1), m.group(1))
    return None


_current_string_re = LazyReCompile(
    '''(?P<open>(?:""")|"|(?:''\')|')(?:((?P<closed>.+?)(?P=open))|'''
    """(?P<unclosed>.+))"""
)


def current_string(cursor_offset: int, line: str) -> LinePart | None:
    """If inside a string of nonzero length, return the string (excluding
    quotes)

    Weaker than bpython.Repl's current_string, because that checks that a
    string is a string based on previous lines in the buffer."""
    for m in _current_string_re.finditer(line):
        i = 3 if m.group(3) else 4
        if m.start(i) <= cursor_offset <= m.end(i):
            return LinePart(m.start(i), m.end(i), m.group(i))
    return None


_current_object_re = LazyReCompile(r"([\w_][\w0-9_]*)[.]")


def current_object(cursor_offset: int, line: str) -> LinePart | None:
    """If in attribute completion, the object on which attribute should be
    looked up."""
    match = current_word(cursor_offset, line)
    if match is None:
        return None
    s = ".".join(
        m.group(1)
        for m in _current_object_re.finditer(match.word)
        if m.end(1) + match.start < cursor_offset
    )
    if not s:
        return None
    return LinePart(match.start, match.start + len(s), s)


_current_object_attribute_re = LazyReCompile(r"([\w_][\w0-9_]*)[.]?")


def current_object_attribute(cursor_offset: int, line: str) -> LinePart | None:
    """If in attribute completion, the attribute being completed"""
    # TODO replace with more general current_expression_attribute
    match = current_word(cursor_offset, line)
    if match is None:
        return None
    matches = _current_object_attribute_re.finditer(match.word)
    next(matches)
    for m in matches:
        if m.start(1) + match.start <= cursor_offset <= m.end(1) + match.start:
            return LinePart(
                m.start(1) + match.start, m.end(1) + match.start, m.group(1)
            )
    return None


_current_from_import_from_re = LazyReCompile(
    r"from +([\w0-9_.]*)(?:\s+import\s+([\w0-9_]+[,]?\s*)+)*"
)


def current_from_import_from(cursor_offset: int, line: str) -> LinePart | None:
    """If in from import completion, the word after from

    returns None if cursor not in or just after one of the two interesting
    parts of an import: from (module) import (name1, name2)
    """
    # TODO allow for as's
    for m in _current_from_import_from_re.finditer(line):
        if (m.start(1) < cursor_offset <= m.end(1)) or (
            m.start(2) < cursor_offset <= m.end(2)
        ):
            return LinePart(m.start(1), m.end(1), m.group(1))
    return None


_current_from_import_import_re_1 = LazyReCompile(
    r"from\s+([\w0-9_.]*)\s+import"
)
_current_from_import_import_re_2 = LazyReCompile(r"([\w0-9_]+)")
_current_from_import_import_re_3 = LazyReCompile(r", *([\w0-9_]*)")


def current_from_import_import(
    cursor_offset: int, line: str
) -> LinePart | None:
    """If in from import completion, the word after import being completed

    returns None if cursor not in or just after one of these words
    """
    baseline = _current_from_import_import_re_1.search(line)
    if baseline is None:
        return None
    match1 = _current_from_import_import_re_2.search(line[baseline.end() :])
    if match1 is None:
        return None
    for m in chain(
        (match1,),
        _current_from_import_import_re_3.finditer(line[baseline.end() :]),
    ):
        start = baseline.end() + m.start(1)
        end = baseline.end() + m.end(1)
        if start < cursor_offset <= end:
            return LinePart(start, end, m.group(1))
    return None


_current_import_re_1 = LazyReCompile(r"import")
_current_import_re_2 = LazyReCompile(r"([\w0-9_.]+)")
_current_import_re_3 = LazyReCompile(r"[,][ ]*([\w0-9_.]*)")


def current_import(cursor_offset: int, line: str) -> LinePart | None:
    # TODO allow for multiple as's
    baseline = _current_import_re_1.search(line)
    if baseline is None:
        return None
    match1 = _current_import_re_2.search(line[baseline.end() :])
    if match1 is None:
        return None
    for m in chain(
        (match1,), _current_import_re_3.finditer(line[baseline.end() :])
    ):
        start = baseline.end() + m.start(1)
        end = baseline.end() + m.end(1)
        if start < cursor_offset <= end:
            return LinePart(start, end, m.group(1))
    return None


_current_method_definition_name_re = LazyReCompile(r"def\s+([a-zA-Z_][\w]*)")


def current_method_definition_name(
    cursor_offset: int, line: str
) -> LinePart | None:
    """The name of a method being defined"""
    for m in _current_method_definition_name_re.finditer(line):
        if m.start(1) <= cursor_offset <= m.end(1):
            return LinePart(m.start(1), m.end(1), m.group(1))
    return None


_current_single_word_re = LazyReCompile(r"(?<![.])\b([a-zA-Z_][\w]*)")


def current_single_word(cursor_offset: int, line: str) -> LinePart | None:
    """the un-dotted word just before or under the cursor"""
    for m in _current_single_word_re.finditer(line):
        if m.start(1) <= cursor_offset <= m.end(1):
            return LinePart(m.start(1), m.end(1), m.group(1))
    return None


def current_dotted_attribute(cursor_offset: int, line: str) -> LinePart | None:
    """The dotted attribute-object pair before the cursor"""
    match = current_word(cursor_offset, line)
    if match is not None and "." in match.word[1:]:
        return match
    return None


_current_expression_attribute_re = LazyReCompile(
    r"[.]\s*((?:[\w_][\w0-9_]*)|(?:))"
)


def current_expression_attribute(
    cursor_offset: int, line: str
) -> LinePart | None:
    """If after a dot, the attribute being completed"""
    # TODO replace with more general current_expression_attribute
    for m in _current_expression_attribute_re.finditer(line):
        if m.start(1) <= cursor_offset <= m.end(1):
            return LinePart(m.start(1), m.end(1), m.group(1))
    return None


def cursor_on_closing_char_pair(
    cursor_offset: int, line: str, ch: str | None = None
) -> tuple[bool, bool]:
    """Checks if cursor sits on closing character of a pair
    and whether its pair character is directly behind it
    """
    on_closing_char, pair_close = False, False
    if line is None:
        return on_closing_char, pair_close
    if cursor_offset < len(line):
        cur_char = line[cursor_offset]
        if cur_char in CHARACTER_PAIR_MAP.values():
            on_closing_char = True if ch is None else cur_char == ch
        if cursor_offset > 0:
            prev_char = line[cursor_offset - 1]
            if (
                on_closing_char
                and prev_char in CHARACTER_PAIR_MAP
                and CHARACTER_PAIR_MAP[prev_char] == cur_char
            ):
                pair_close = True if ch is None else prev_char == ch
    return on_closing_char, pair_close



# The MIT License
#
# Copyright (c) 2009-2015 the bpython authors.
# Copyright (c) 2015-2020 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# To gradually migrate to mypy we aren't setting these globally yet
# mypy: disallow_untyped_defs=True
# mypy: disallow_untyped_calls=True

import __main__
import abc
import glob
import itertools
import keyword
import logging
import os
import re
import rlcompleter
import builtins

from enum import Enum
from typing import (
    Any,
    Optional,
)
from collections.abc import Iterator, Sequence

from . import inspection
from . import line as lineparts
from .line import LinePart
from .lazyre import LazyReCompile
from .simpleeval import safe_eval, evaluate_current_expression, EvaluationError
from .importcompletion import ModuleGatherer


logger = logging.getLogger(__name__)


# Autocomplete modes
class AutocompleteModes(Enum):
    NONE = "none"
    SIMPLE = "simple"
    SUBSTRING = "substring"
    FUZZY = "fuzzy"

    @classmethod
    def from_string(cls, value: str) -> Optional["AutocompleteModes"]:
        if value.upper() in cls.__members__:
            return cls.__members__[value.upper()]
        return None


MAGIC_METHODS = tuple(
    f"__{s}__"
    for s in (
        "new",
        "init",
        "del",
        "repr",
        "str",
        "bytes",
        "format",
        "lt",
        "le",
        "eq",
        "ne",
        "gt",
        "ge",
        "hash",
        "bool",
        "getattr",
        "getattribute",
        "setattr",
        "delattr",
        "dir",
        "get",
        "set",
        "delete",
        "set_name",
        "init_subclass",
        "instancecheck",
        "subclasscheck",
        "class_getitem",
        "call",
        "len",
        "length_hint",
        "getitem",
        "setitem",
        "delitem",
        "missing",
        "iter",
        "reversed",
        "contains",
        "add",
        "sub",
        "mul",
        "matmul",
        "truediv",
        "floordiv",
        "mod",
        "divmod",
        "pow",
        "lshift",
        "rshift",
        "and",
        "xor",
        "or",
        "radd",
        "rsub",
        "rmul",
        "rmatmul",
        "rtruediv",
        "rfloordiv",
        "rmod",
        "rdivmod",
        "rpow",
        "rlshift",
        "rrshift",
        "rand",
        "rxor",
        "ror",
        "iadd",
        "isub",
        "imul",
        "imatmul",
        "itruediv",
        "ifloordiv",
        "imod",
        "ipow",
        "ilshift",
        "irshift",
        "iand",
        "ixor",
        "ixor",
        "neg",
        "pos",
        "abs",
        "invert",
        "complex",
        "int",
        "float",
        "index",
        "round",
        "trunc",
        "floor",
        "ceil",
        "enter",
        "exit",
        "await",
        "aiter",
        "anext",
        "aenter",
        "aexit",
    )
)

KEYWORDS = frozenset(keyword.kwlist)


def _after_last_dot(name: str) -> str:
    return name.rstrip(".").rsplit(".")[-1]


def _few_enough_underscores(current: str, match: str) -> bool:
    """Returns whether match should be shown based on current

    if current is _, True if match starts with 0 or 1 underscore
    if current is __, True regardless of match
    otherwise True if match does not start with any underscore
    """
    if current.startswith("__"):
        return True
    elif current.startswith("_") and not match.startswith("__"):
        return True
    return not match.startswith("_")


def _method_match_none(word: str, size: int, text: str) -> bool:
    return False


def _method_match_simple(word: str, size: int, text: str) -> bool:
    return word[:size] == text


def _method_match_substring(word: str, size: int, text: str) -> bool:
    return text in word


def _method_match_fuzzy(word: str, size: int, text: str) -> bool:
    s = r".*{}.*".format(".*".join(c for c in text))
    return re.search(s, word) is not None


_MODES_MAP = {
    AutocompleteModes.NONE: _method_match_none,
    AutocompleteModes.SIMPLE: _method_match_simple,
    AutocompleteModes.SUBSTRING: _method_match_substring,
    AutocompleteModes.FUZZY: _method_match_fuzzy,
}


class BaseCompletionType:
    """Describes different completion types"""

    def __init__(
        self,
        shown_before_tab: bool = True,
        mode: AutocompleteModes = AutocompleteModes.SIMPLE,
    ) -> None:
        self._shown_before_tab = shown_before_tab
        self.method_match = _MODES_MAP[mode]

    @abc.abstractmethod
    def matches(
        self, cursor_offset: int, line: str, **kwargs: Any
    ) -> set[str] | None:
        """Returns a list of possible matches given a line and cursor, or None
        if this completion type isn't applicable.

        ie, import completion doesn't make sense if there cursor isn't after
        an import or from statement, so it ought to return None.

        Completion types are used to:
            * `locate(cur, line)` their initial target word to replace given a
              line and cursor
            * find `matches(cur, line)` that might replace that word
            * `format(match)` matches to be displayed to the user
            * determine whether suggestions should be `shown_before_tab`
            * `substitute(cur, line, match)` in a match for what's found with
              `target`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        """Returns a Linepart namedtuple instance or None given cursor and line

        A Linepart namedtuple contains a start, stop, and word. None is
        returned if no target for this type of completion is found under
        the cursor."""
        raise NotImplementedError

    def format(self, word: str) -> str:
        return word

    def substitute(
        self, cursor_offset: int, line: str, match: str
    ) -> tuple[int, str]:
        """Returns a cursor offset and line with match swapped in"""
        lpart = self.locate(cursor_offset, line)
        assert lpart
        offset = lpart.start + len(match)
        changed_line = line[: lpart.start] + match + line[lpart.stop :]
        return offset, changed_line

    @property
    def shown_before_tab(self) -> bool:
        """Whether suggestions should be shown before the user hits tab, or only
        once that has happened."""
        return self._shown_before_tab


class CumulativeCompleter(BaseCompletionType):
    """Returns combined matches from several completers"""

    def __init__(
        self,
        completers: Sequence[BaseCompletionType],
        mode: AutocompleteModes = AutocompleteModes.SIMPLE,
    ) -> None:
        if not completers:
            raise ValueError(
                "CumulativeCompleter requires at least one completer"
            )
        self._completers: Sequence[BaseCompletionType] = completers

        super().__init__(True, mode)

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        for completer in self._completers:
            return_value = completer.locate(cursor_offset, line)
            if return_value is not None:
                return return_value
        return None

    def format(self, word: str) -> str:
        return self._completers[0].format(word)

    def matches(
        self, cursor_offset: int, line: str, **kwargs: Any
    ) -> set[str] | None:
        return_value = None
        all_matches = set()
        for completer in self._completers:
            matches = completer.matches(
                cursor_offset=cursor_offset, line=line, **kwargs
            )
            if matches is not None:
                all_matches.update(matches)
                return_value = all_matches

        return return_value


class ImportCompletion(BaseCompletionType):
    def __init__(
        self,
        module_gatherer: ModuleGatherer,
        mode: AutocompleteModes = AutocompleteModes.SIMPLE,
    ):
        super().__init__(False, mode)
        self.module_gatherer = module_gatherer

    def matches(
        self, cursor_offset: int, line: str, **kwargs: Any
    ) -> set[str] | None:
        return self.module_gatherer.complete(cursor_offset, line)

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_word(cursor_offset, line)

    def format(self, word: str) -> str:
        return _after_last_dot(word)


def _safe_glob(pathname: str) -> Iterator[str]:
    return glob.iglob(glob.escape(pathname) + "*")


class FilenameCompletion(BaseCompletionType):
    def __init__(self, mode: AutocompleteModes = AutocompleteModes.SIMPLE):
        super().__init__(False, mode)

    def matches(
        self, cursor_offset: int, line: str, **kwargs: Any
    ) -> set[str] | None:
        cs = lineparts.current_string(cursor_offset, line)
        if cs is None:
            return None
        matches = set()
        username = cs.word.split(os.path.sep, 1)[0]
        user_dir = os.path.expanduser(username)
        for filename in _safe_glob(os.path.expanduser(cs.word)):
            if os.path.isdir(filename):
                filename += os.path.sep
            if cs.word.startswith("~"):
                filename = username + filename[len(user_dir) :]
            matches.add(filename)
        return matches

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_string(cursor_offset, line)

    def format(self, filename: str) -> str:
        if os.sep in filename[:-1]:
            return filename[filename.rindex(os.sep, 0, -1) + 1 :]
        else:
            return filename


class AttrCompletion(BaseCompletionType):
    attr_matches_re = LazyReCompile(r"(\w+(\.\w+)*)\.(\w*)")

    def matches(
        self,
        cursor_offset: int,
        line: str,
        *,
        locals_: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> set[str] | None:
        r = self.locate(cursor_offset, line)
        if r is None:
            return None

        if locals_ is None:  # TODO add a note about why
            locals_ = __main__.__dict__

        assert "." in r.word

        i = r.word.rfind("[") + 1
        methodtext = r.word[i:]
        matches = {
            "".join([r.word[:i], m])
            for m in self.attr_matches(methodtext, locals_)
        }

        return {
            m
            for m in matches
            if _few_enough_underscores(r.word.split(".")[-1], m.split(".")[-1])
        }

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_dotted_attribute(cursor_offset, line)

    def format(self, word: str) -> str:
        return _after_last_dot(word)

    def attr_matches(
        self, text: str, namespace: dict[str, Any]
    ) -> Iterator[str]:
        """Taken from rlcompleter.py and bent to my will."""

        m = self.attr_matches_re.match(text)
        if not m:
            return (_ for _ in ())

        expr, attr = m.group(1, 3)
        if expr.isdigit():
            # Special case: float literal, using attrs here will result in
            # a SyntaxError
            return (_ for _ in ())
        try:
            obj = safe_eval(expr, namespace)
        except EvaluationError:
            return (_ for _ in ())
        return self.attr_lookup(obj, expr, attr)

    def attr_lookup(self, obj: Any, expr: str, attr: str) -> Iterator[str]:
        """Second half of attr_matches."""
        words = self.list_attributes(obj)
        if inspection.hasattr_safe(obj, "__class__"):
            words.append("__class__")
            klass = inspection.getattr_safe(obj, "__class__")
            words = words + rlcompleter.get_class_members(klass)
            if not isinstance(klass, abc.ABCMeta):
                try:
                    words.remove("__abstractmethods__")
                except ValueError:
                    pass

        n = len(attr)
        return (
            f"{expr}.{word}"
            for word in words
            if self.method_match(word, n, attr) and word != "__builtins__"
        )

    def list_attributes(self, obj: Any) -> list[str]:
        # TODO: re-implement dir without AttrCleaner here
        #
        # Note: accessing `obj.__dir__` via `getattr_static` is not side-effect free.
        with inspection.AttrCleaner(obj):
            return dir(obj)


class DictKeyCompletion(BaseCompletionType):
    def matches(
        self,
        cursor_offset: int,
        line: str,
        *,
        locals_: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> set[str] | None:
        if locals_ is None:
            return None

        r = self.locate(cursor_offset, line)
        if r is None:
            return None
        current_dict_parts = lineparts.current_dict(cursor_offset, line)
        if current_dict_parts is None:
            return None

        dexpr = current_dict_parts.word
        try:
            obj = safe_eval(dexpr, locals_)
        except EvaluationError:
            return None
        if isinstance(obj, dict) and obj.keys():
            matches = {
                f"{k!r}]" for k in obj.keys() if repr(k).startswith(r.word)
            }
            return matches if matches else None
        else:
            return None

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_dict_key(cursor_offset, line)

    def format(self, match: str) -> str:
        return match[:-1]


class MagicMethodCompletion(BaseCompletionType):
    def matches(
        self,
        cursor_offset: int,
        line: str,
        *,
        current_block: str | None = None,
        complete_magic_methods: bool | None = None,
        **kwargs: Any,
    ) -> set[str] | None:
        if (
            current_block is None
            or complete_magic_methods is None
            or not complete_magic_methods
        ):
            return None

        r = self.locate(cursor_offset, line)
        if r is None:
            return None
        if "class" not in current_block:
            return None
        return {name for name in MAGIC_METHODS if name.startswith(r.word)}

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_method_definition_name(cursor_offset, line)


class GlobalCompletion(BaseCompletionType):
    def matches(
        self,
        cursor_offset: int,
        line: str,
        *,
        locals_: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> set[str] | None:
        """Compute matches when text is a simple name.
        Return a list of all keywords, built-in functions and names currently
        defined in self.namespace that match.
        """
        if locals_ is None:
            return None

        r = self.locate(cursor_offset, line)
        if r is None:
            return None

        n = len(r.word)
        matches = {
            word for word in KEYWORDS if self.method_match(word, n, r.word)
        }
        for nspace in (builtins.__dict__, locals_):
            for word, val in nspace.items():
                # if identifier isn't ascii, don't complete (syntax error)
                if word is None:
                    continue
                if (
                    self.method_match(word, n, r.word)
                    and word != "__builtins__"
                ):
                    matches.add(_callable_postfix(val, word))
        return matches if matches else None

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_single_word(cursor_offset, line)


class ParameterNameCompletion(BaseCompletionType):
    def matches(
        self,
        cursor_offset: int,
        line: str,
        *,
        funcprops: inspection.FuncProps | None = None,
        **kwargs: Any,
    ) -> set[str] | None:
        if funcprops is None:
            return None

        r = self.locate(cursor_offset, line)
        if r is None:
            return None

        matches = {
            f"{name}="
            for name in funcprops.argspec.args
            if isinstance(name, str) and name.startswith(r.word)
        }
        matches.update(
            f"{name}="
            for name in funcprops.argspec.kwonly
            if name.startswith(r.word)
        )
        return matches if matches else None

    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        r = lineparts.current_word(cursor_offset, line)
        if r and r.word[-1] == "(":
            # if the word ends with a (, it's the parent word with an empty
            # param. Return an empty word
            return lineparts.LinePart(r.stop, r.stop, "")
        return r


class ExpressionAttributeCompletion(AttrCompletion):
    # could replace attr completion as a more general case with some work
    def locate(self, cursor_offset: int, line: str) -> LinePart | None:
        return lineparts.current_expression_attribute(cursor_offset, line)

    def matches(
        self,
        cursor_offset: int,
        line: str,
        *,
        locals_: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> set[str] | None:
        if locals_ is None:
            locals_ = __main__.__dict__

        attr = self.locate(cursor_offset, line)
        assert attr, "locate was already truthy for the same call"

        try:
            obj = evaluate_current_expression(cursor_offset, line, locals_)
        except EvaluationError:
            return set()

        # strips leading dot
        matches = (m[1:] for m in self.attr_lookup(obj, "", attr.word))
        return {m for m in matches if _few_enough_underscores(attr.word, m)}


try:
    import jedi
except ImportError:

    class MultilineJediCompletion(BaseCompletionType):  # type: ignore [no-redef]
        def matches(
            self, cursor_offset: int, line: str, **kwargs: Any
        ) -> set[str] | None:
            return None

        def locate(self, cursor_offset: int, line: str) -> LinePart | None:
            return None

else:

    class MultilineJediCompletion(BaseCompletionType):  # type: ignore [no-redef]
        _orig_start: int | None

        def matches(
            self,
            cursor_offset: int,
            line: str,
            *,
            current_block: str | None = None,
            history: list[str] | None = None,
            **kwargs: Any,
        ) -> set[str] | None:
            if (
                current_block is None
                or history is None
                or "\n" not in current_block
                or not lineparts.current_word(cursor_offset, line)
            ):
                return None

            assert cursor_offset <= len(line), "{!r} {!r}".format(
                cursor_offset,
                line,
            )

            combined_history = "\n".join(itertools.chain(history, (line,)))
            try:
                script = jedi.Script(combined_history, path="fake.py")
                completions = script.complete(
                    combined_history.count("\n") + 1, cursor_offset
                )
            except (jedi.NotFoundError, IndexError, KeyError):
                # IndexError for #483
                # KeyError for #544
                self._orig_start = None
                return None

            if completions:
                diff = len(completions[0].name) - len(completions[0].complete)
                self._orig_start = cursor_offset - diff
            else:
                self._orig_start = None
                return None
            assert isinstance(self._orig_start, int)

            matches = [c.name for c in completions]
            if any(
                not m.lower().startswith(matches[0][0].lower()) for m in matches
            ):
                # Too general - giving completions starting with multiple
                # letters
                return None
            else:
                # case-sensitive matches only
                first_letter = line[self._orig_start]
                return {m for m in matches if m.startswith(first_letter)}

        def locate(self, cursor_offset: int, line: str) -> LinePart:
            assert self._orig_start is not None
            start = self._orig_start
            end = cursor_offset
            return LinePart(start, end, line[start:end])


def get_completer(
    completers: Sequence[BaseCompletionType],
    cursor_offset: int,
    line: str,
    *,
    locals_: dict[str, Any] | None = None,
    argspec: inspection.FuncProps | None = None,
    history: list[str] | None = None,
    current_block: str | None = None,
    complete_magic_methods: bool | None = None,
) -> tuple[list[str], BaseCompletionType | None]:
    """Returns a list of matches and an applicable completer

    If no matches available, returns a tuple of an empty list and None

    cursor_offset is the current cursor column
    line is a string of the current line
    kwargs (all optional):
        locals_ is a dictionary of the environment
        argspec is an inspection.FuncProps instance for the current function where
            the cursor is
        current_block is the possibly multiline not-yet-evaluated block of
            code which the current line is part of
        complete_magic_methods is a bool of whether we ought to complete
            double underscore methods like __len__ in method signatures
    """

    def _cmpl_sort(x: str) -> tuple[bool, str]:
        """
        Function used to sort the matches.
        """
        # put parameters above everything in completion
        return (
            x[-1] != "=",
            x,
        )

    for completer in completers:
        try:
            matches = completer.matches(
                cursor_offset,
                line,
                locals_=locals_,
                funcprops=argspec,
                history=history,
                current_block=current_block,
                complete_magic_methods=complete_magic_methods,
            )
        except Exception as e:
            # Instead of crashing the UI, log exceptions from autocompleters.
            logger.debug(
                "Completer %r failed with unhandled exception: %s", completer, e
            )
            continue
        if matches is not None:
            return sorted(matches, key=_cmpl_sort), (
                completer if matches else None
            )

    return [], None


def get_default_completer(
    mode: AutocompleteModes, module_gatherer: ModuleGatherer
) -> tuple[BaseCompletionType, ...]:
    return (
        (
            DictKeyCompletion(mode=mode),
            ImportCompletion(module_gatherer, mode=mode),
            FilenameCompletion(mode=mode),
            MagicMethodCompletion(mode=mode),
            MultilineJediCompletion(mode=mode),
            CumulativeCompleter(
                (
                    GlobalCompletion(mode=mode),
                    ParameterNameCompletion(mode=mode),
                ),
                mode=mode,
            ),
            AttrCompletion(mode=mode),
            ExpressionAttributeCompletion(mode=mode),
        )
        if mode != AutocompleteModes.NONE
        else tuple()
    )


def _callable_postfix(value: Any, word: str) -> str:
    """rlcompleter's _callable_postfix done right."""
    if callable(value):
        word += "("
    return word



# The MIT License
#
# Copyright (c) 2014-2022 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import subprocess
from typing import Protocol
from urllib.parse import urljoin, urlparse

import requests
import unicodedata

from .config import getpreferredencoding
from .translations import _


class PasteFailed(Exception):
    pass


class Paster(Protocol):
    def paste(self, s: str) -> tuple[str, str | None]: ...


class PastePinnwand:
    def __init__(self, url: str, expiry: str) -> None:
        self.url = url
        self.expiry = expiry

    def paste(self, s: str) -> tuple[str, str]:
        """Upload to pastebin via json interface."""

        url = urljoin(self.url, "/api/v1/paste")
        payload = {
            "expiry": self.expiry,
            "files": [{"lexer": "pycon", "content": s}],
        }

        try:
            response = requests.post(url, json=payload, verify=True)
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            raise PasteFailed(str(exc))

        data = response.json()

        paste_url = data["link"]
        removal_url = data["removal"]

        return (paste_url, removal_url)


class PasteHelper:
    def __init__(self, executable: str) -> None:
        self.executable = executable

    def paste(self, s: str) -> tuple[str, None]:
        """Call out to helper program for pastebin upload."""

        try:
            helper = subprocess.Popen(
                "",
                executable=self.executable,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            assert helper.stdin is not None
            encoding = getpreferredencoding()
            helper.stdin.write(s.encode(encoding))
            output = helper.communicate()[0].decode(encoding)
            paste_url = output.split()[0]
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise PasteFailed(_("Helper program not found."))
            else:
                raise PasteFailed(_("Helper program could not be run."))

        if helper.returncode != 0:
            raise PasteFailed(
                _(
                    "Helper program returned non-zero exit status %d."
                    % (helper.returncode,)
                )
            )

        if not paste_url:
            raise PasteFailed(_("No output from helper program."))

        parsed_url = urlparse(paste_url)
        if not parsed_url.scheme or any(
            unicodedata.category(c) == "Cc" for c in paste_url
        ):
            raise PasteFailed(
                _(
                    "Failed to recognize the helper "
                    "program's output as an URL."
                )
            )

        return paste_url, None



import logging
import itertools

from curtsies import fsarray, fmtstr, FSArray
from curtsies.formatstring import linesplit
from curtsies.fmtfuncs import bold

from .parse import func_for_letter

logger = logging.getLogger(__name__)

# All paint functions should
# * return an array of the width they were asked for
# * return an array not taller than the height they were asked for


def display_linize(msg, columns, blank_line=False):
    """Returns lines obtained by splitting msg over multiple lines.

    Warning: if msg is empty, returns an empty list of lines"""
    if not msg:
        return [""] if blank_line else []
    msg = fmtstr(msg)
    try:
        display_lines = list(msg.width_aware_splitlines(columns))
    # use old method if wcwidth can't determine width of msg
    except ValueError:
        display_lines = [
            msg[start:end]
            for start, end in zip(
                range(0, len(msg), columns),
                range(columns, len(msg) + columns, columns),
            )
        ]
    return display_lines


def paint_history(rows, columns, display_lines):
    lines = []
    for r, line in zip(range(rows), display_lines[-rows:]):
        lines.append(fmtstr(line[:columns]))
    r = fsarray(lines, width=columns)
    assert r.shape[0] <= rows, repr(r.shape) + " " + repr(rows)
    assert r.shape[1] <= columns, repr(r.shape) + " " + repr(columns)
    return r


def paint_current_line(rows, columns, current_display_line):
    lines = display_linize(current_display_line, columns, True)
    return fsarray(lines, width=columns)


def paginate(rows, matches, current, words_wide):
    if current not in matches:
        current = matches[0]
    per_page = rows * words_wide
    current_page = matches.index(current) // per_page
    return matches[per_page * current_page : per_page * (current_page + 1)]


def matches_lines(rows, columns, matches, current, config, match_format):
    highlight_color = func_for_letter(config.color_scheme["operator"].lower())

    if not matches:
        return []
    color = func_for_letter(config.color_scheme["main"])
    max_match_width = max(len(m) for m in matches)
    words_wide = max(1, (columns - 1) // (max_match_width + 1))
    matches = [match_format(m) for m in matches]
    if current:
        current = match_format(current)

    matches = paginate(rows, matches, current, words_wide)

    result = [
        fmtstr(" ").join(
            (
                color(m.ljust(max_match_width))
                if m != current
                else highlight_color(m.ljust(max_match_width))
            )
            for m in matches[i : i + words_wide]
        )
        for i in range(0, len(matches), words_wide)
    ]

    logger.debug("match: %r" % current)
    logger.debug("matches_lines: %r" % result)
    return result


def formatted_argspec(funcprops, arg_pos, columns, config):
    # Pretty directly taken from bpython.cli
    func = funcprops.func
    args = funcprops.argspec.args
    kwargs = funcprops.argspec.defaults
    _args = funcprops.argspec.varargs
    _kwargs = funcprops.argspec.varkwargs
    is_bound_method = funcprops.is_bound_method
    kwonly = funcprops.argspec.kwonly
    kwonly_defaults = funcprops.argspec.kwonly_defaults or dict()

    arg_color = func_for_letter(config.color_scheme["name"])
    func_color = func_for_letter(config.color_scheme["name"].swapcase())
    punctuation_color = func_for_letter(config.color_scheme["punctuation"])
    token_color = func_for_letter(config.color_scheme["token"])
    bolds = {
        token_color: lambda x: bold(token_color(x)),
        arg_color: lambda x: bold(arg_color(x)),
    }

    s = func_color(func) + arg_color(": (")

    if is_bound_method and isinstance(arg_pos, int):
        # TODO what values could this have?
        arg_pos += 1

    for i, arg in enumerate(args):
        kw = None
        if kwargs and i >= len(args) - len(kwargs):
            kw = str(kwargs[i - (len(args) - len(kwargs))])
        color = token_color if arg_pos in (i, arg) else arg_color
        if i == arg_pos or arg == arg_pos:
            color = bolds[color]

        s += color(arg)

        if kw is not None:
            s += punctuation_color("=")
            s += token_color(kw)

        if i != len(args) - 1:
            s += punctuation_color(", ")

    if _args:
        if args:
            s += punctuation_color(", ")
        s += token_color(f"*{_args}")

    if kwonly:
        if not _args:
            if args:
                s += punctuation_color(", ")
            s += punctuation_color("*")
        marker = object()
        for arg in kwonly:
            s += punctuation_color(", ")
            color = token_color
            if arg_pos:
                color = bolds[color]
            s += color(arg)
            default = kwonly_defaults.get(arg, marker)
            if default is not marker:
                s += punctuation_color("=")
                s += token_color(repr(default))

    if _kwargs:
        if args or _args or kwonly:
            s += punctuation_color(", ")
        s += token_color(f"**{_kwargs}")
    s += punctuation_color(")")

    return linesplit(s, columns)


def formatted_docstring(docstring, columns, config):
    if isinstance(docstring, bytes):
        docstring = docstring.decode("utf8")
    elif isinstance(docstring, str):
        pass
    else:
        # TODO: fail properly here and catch possible exceptions in callers.
        return []
    color = func_for_letter(config.color_scheme["comment"])
    return sum(
        (
            [
                color(x)
                for x in (display_linize(line, columns) if line else fmtstr(""))
            ]
            for line in docstring.split("\n")
        ),
        [],
    )


def paint_infobox(
    rows,
    columns,
    matches,
    funcprops,
    arg_pos,
    match,
    docstring,
    config,
    match_format,
):
    """Returns painted completions, funcprops, match, docstring etc."""
    if not (rows and columns):
        return FSArray(0, 0)
    width = columns - 4
    from_argspec = (
        formatted_argspec(funcprops, arg_pos, width, config)
        if funcprops
        else []
    )
    from_doc = (
        formatted_docstring(docstring, width, config) if docstring else []
    )
    from_matches = (
        matches_lines(
            max(1, rows - len(from_argspec) - 2),
            width,
            matches,
            match,
            config,
            match_format,
        )
        if matches
        else []
    )

    lines = from_argspec + from_matches + from_doc

    def add_border(line):
        """Add colored borders left and right to a line."""
        new_line = border_color(config.left_border + " ")
        new_line += line.ljust(width)[:width]
        new_line += border_color(" " + config.right_border)
        return new_line

    border_color = func_for_letter(config.color_scheme["main"])

    top_line = border_color(
        config.left_top_corner
        + config.top_border * (width + 2)
        + config.right_top_corner
    )
    bottom_line = border_color(
        config.left_bottom_corner
        + config.bottom_border * (width + 2)
        + config.right_bottom_corner
    )

    output_lines = list(
        itertools.chain((top_line,), map(add_border, lines), (bottom_line,))
    )
    r = fsarray(
        output_lines[: min(rows - 1, len(output_lines) - 1)] + output_lines[-1:]
    )
    return r


def paint_last_events(rows, columns, names, config):
    if not names:
        return fsarray([])
    width = min(max(len(name) for name in names), columns - 2)
    output_lines = []
    output_lines.append(
        config.left_top_corner
        + config.top_border * width
        + config.right_top_corner
    )
    for name in reversed(names[max(0, len(names) - (rows - 2)) :]):
        output_lines.append(
            config.left_border
            + name[:width].center(width)
            + config.right_border
        )
    output_lines.append(
        config.left_bottom_corner
        + config.bottom_border * width
        + config.right_bottom_corner
    )
    return fsarray(output_lines)


def paint_statusbar(rows, columns, msg, config):
    func = func_for_letter(config.color_scheme["main"])
    return fsarray([func(msg.ljust(columns))[:columns]])



# The MIT License
#
# Copyright (c) 2009 the bpython authors.
# Copyright (c) 2012-2021 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
from pathlib import Path
import stat
from itertools import islice, chain
from typing import TextIO
from collections.abc import Iterable

from .translations import _
from .filelock import FileLock


class History:
    """Stores readline-style history and current place in it"""

    def __init__(
        self,
        entries: Iterable[str] | None = None,
        duplicates: bool = True,
        hist_size: int = 100,
    ) -> None:
        if entries is None:
            self.entries = [""]
        else:
            self.entries = list(entries)
        # how many lines back in history is currently selected where 0 is the
        # saved typed line, 1 the prev entered line
        self.index = 0
        # what was on the prompt before using history
        self.saved_line = ""
        self.duplicates = duplicates
        self.hist_size = hist_size

    def append(self, line: str) -> None:
        self.append_to(self.entries, line)

    def append_to(self, entries: list[str], line: str) -> None:
        line = line.rstrip("\n")
        if line:
            if not self.duplicates:
                # remove duplicates
                try:
                    while True:
                        entries.remove(line)
                except ValueError:
                    pass
            entries.append(line)

    def first(self) -> str:
        """Move back to the beginning of the history."""
        if not self.is_at_end:
            self.index = len(self.entries)
        return self.entries[-self.index]

    def back(
        self,
        start: bool = True,
        search: bool = False,
        target: str | None = None,
        include_current: bool = False,
    ) -> str:
        """Move one step back in the history."""
        if target is None:
            target = self.saved_line
        if not self.is_at_end:
            if search:
                self.index += self.find_partial_match_backward(
                    target, include_current
                )
            elif start:
                self.index += self.find_match_backward(target, include_current)
            else:
                self.index += 1
        return self.entry

    @property
    def entry(self) -> str:
        """The current entry, which may be the saved line"""
        return self.entries[-self.index] if self.index else self.saved_line

    @property
    def entries_by_index(self) -> list[str]:
        return list(chain((self.saved_line,), reversed(self.entries)))

    def find_match_backward(
        self, search_term: str, include_current: bool = False
    ) -> int:
        add = 0 if include_current else 1
        start = self.index + add
        for idx, val in enumerate(islice(self.entries_by_index, start, None)):
            if val.startswith(search_term):
                return idx + add
        return 0

    def find_partial_match_backward(
        self, search_term: str, include_current: bool = False
    ) -> int:
        add = 0 if include_current else 1
        start = self.index + add
        for idx, val in enumerate(islice(self.entries_by_index, start, None)):
            if search_term in val:
                return idx + add
        return 0

    def forward(
        self,
        start: bool = True,
        search: bool = False,
        target: str | None = None,
        include_current: bool = False,
    ) -> str:
        """Move one step forward in the history."""
        if target is None:
            target = self.saved_line
        if self.index > 1:
            if search:
                self.index -= self.find_partial_match_forward(
                    target, include_current
                )
            elif start:
                self.index -= self.find_match_forward(target, include_current)
            else:
                self.index -= 1
            return self.entry
        else:
            self.index = 0
            return self.saved_line

    def find_match_forward(
        self, search_term: str, include_current: bool = False
    ) -> int:
        add = 0 if include_current else 1
        end = max(0, self.index - (1 - add))
        for idx in range(end):
            val = self.entries_by_index[end - 1 - idx]
            if val.startswith(search_term):
                return idx + (0 if include_current else 1)
        return self.index

    def find_partial_match_forward(
        self, search_term: str, include_current: bool = False
    ) -> int:
        add = 0 if include_current else 1
        end = max(0, self.index - (1 - add))
        for idx in range(end):
            val = self.entries_by_index[end - 1 - idx]
            if search_term in val:
                return idx + add
        return self.index

    def last(self) -> str:
        """Move forward to the end of the history."""
        if not self.is_at_start:
            self.index = 0
        return self.entries[0]

    @property
    def is_at_end(self) -> bool:
        return self.index >= len(self.entries) or self.index == -1

    @property
    def is_at_start(self) -> bool:
        return self.index == 0

    def enter(self, line: str) -> None:
        if self.index == 0:
            self.saved_line = line

    def reset(self) -> None:
        self.index = 0
        self.saved_line = ""

    def load(self, filename: Path, encoding: str) -> None:
        with open(filename, encoding=encoding, errors="ignore") as hfile:
            with FileLock(hfile, filename=str(filename)):
                self.entries = self.load_from(hfile)

    def load_from(self, fd: TextIO) -> list[str]:
        entries: list[str] = []
        for line in fd:
            self.append_to(entries, line)
        return entries if len(entries) else [""]

    def save(self, filename: Path, encoding: str, lines: int = 0) -> None:
        fd = os.open(
            filename,
            os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
            stat.S_IRUSR | stat.S_IWUSR,
        )
        with open(fd, "w", encoding=encoding, errors="ignore") as hfile:
            with FileLock(hfile, filename=str(filename)):
                self.save_to(hfile, self.entries, lines)

    def save_to(
        self, fd: TextIO, entries: list[str] | None = None, lines: int = 0
    ) -> None:
        if entries is None:
            entries = self.entries
        for line in entries[-lines:]:
            fd.write(line)
            fd.write("\n")

    def append_reload_and_write(
        self, s: str, filename: Path, encoding: str
    ) -> None:
        if not self.hist_size:
            return self.append(s)

        try:
            fd = os.open(
                filename,
                os.O_APPEND | os.O_RDWR | os.O_CREAT,
                stat.S_IRUSR | stat.S_IWUSR,
            )
            with open(fd, "a+", encoding=encoding, errors="ignore") as hfile:
                with FileLock(hfile, filename=str(filename)):
                    # read entries
                    hfile.seek(0, os.SEEK_SET)
                    entries = self.load_from(hfile)
                    self.append_to(entries, s)

                    # write new entries
                    hfile.seek(0, os.SEEK_SET)
                    hfile.truncate()
                    self.save_to(hfile, entries, self.hist_size)

                    self.entries = entries
        except OSError as err:
            raise RuntimeError(
                _("Error occurred while writing to file %s (%s)")
                % (filename, err.strerror)
            )
        else:
            if len(self.entries) == 0:
                # Make sure that entries contains at least one element. If the
                # file and s are empty, this can occur.
                self.entries = [""]



# To gradually migrate to mypy we aren't setting these globally yet
# mypy: disallow_untyped_defs=True
# mypy: disallow_untyped_calls=True

import argparse
import collections
import logging
import sys

import curtsies
import curtsies.events
import curtsies.input
import curtsies.window

from . import args as bpargs, translations, inspection
from .config import Config
from .curtsiesfrontend import events
from .curtsiesfrontend.coderunner import SystemExitFromCodeRunner
from .curtsiesfrontend.interpreter import Interp
from .curtsiesfrontend.repl import BaseRepl
from .repl import extract_exit_value
from .translations import _

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Union,
)
from collections.abc import Callable, Generator, Sequence

logger = logging.getLogger(__name__)


class SupportsEventGeneration(Protocol):
    def send(
        self, timeout: float | None
    ) -> str | curtsies.events.Event | None: ...

    def __iter__(self) -> "SupportsEventGeneration": ...

    def __next__(self) -> str | curtsies.events.Event | None: ...


class FullCurtsiesRepl(BaseRepl):
    def __init__(
        self,
        config: Config,
        locals_: dict[str, Any] | None = None,
        banner: str | None = None,
        interp: Interp | None = None,
    ) -> None:
        self.input_generator = curtsies.input.Input(
            keynames="curtsies", sigint_event=True, paste_threshold=None
        )
        window = curtsies.window.CursorAwareWindow(
            sys.stdout,
            sys.stdin,
            keep_last_line=True,
            hide_cursor=False,
            extra_bytes_callback=self.input_generator.unget_bytes,
        )

        self._request_refresh_callback: Callable[[], None] = (
            self.input_generator.event_trigger(events.RefreshRequestEvent)
        )
        self._schedule_refresh_callback = (
            self.input_generator.scheduled_event_trigger(
                events.ScheduledRefreshRequestEvent
            )
        )
        self._request_reload_callback = (
            self.input_generator.threadsafe_event_trigger(events.ReloadEvent)
        )
        self._interrupting_refresh_callback = (
            self.input_generator.threadsafe_event_trigger(lambda: None)
        )
        self._request_undo_callback = self.input_generator.event_trigger(
            events.UndoEvent
        )

        with self.input_generator:
            pass  # temp hack to get .original_stty

        super().__init__(
            config,
            window,
            locals_=locals_,
            banner=banner,
            interp=interp,
            orig_tcattrs=self.input_generator.original_stty,
        )

    def _request_refresh(self) -> None:
        return self._request_refresh_callback()

    def _schedule_refresh(self, when: float) -> None:
        return self._schedule_refresh_callback(when)

    def _request_reload(self, files_modified: Sequence[str]) -> None:
        return self._request_reload_callback(files_modified=files_modified)

    def interrupting_refresh(self) -> None:
        return self._interrupting_refresh_callback()

    def request_undo(self, n: int = 1) -> None:
        return self._request_undo_callback(n=n)

    def get_term_hw(self) -> tuple[int, int]:
        return self.window.get_term_hw()

    def get_cursor_vertical_diff(self) -> int:
        return self.window.get_cursor_vertical_diff()

    def get_top_usable_line(self) -> int:
        return self.window.top_usable_row

    def on_suspend(self) -> None:
        self.window.__exit__(None, None, None)
        self.input_generator.__exit__(None, None, None)

    def after_suspend(self) -> None:
        self.input_generator.__enter__()
        self.window.__enter__()
        self.interrupting_refresh()

    def process_event_and_paint(
        self, e: str | curtsies.events.Event | None
    ) -> None:
        """If None is passed in, just paint the screen"""
        try:
            if e is not None:
                self.process_event(e)
        except (SystemExitFromCodeRunner, SystemExit) as err:
            array, cursor_pos = self.paint(
                about_to_exit=True,
                user_quit=isinstance(err, SystemExitFromCodeRunner),
            )
            scrolled = self.window.render_to_terminal(array, cursor_pos)
            self.scroll_offset += scrolled
            raise
        else:
            array, cursor_pos = self.paint()
            scrolled = self.window.render_to_terminal(array, cursor_pos)
            self.scroll_offset += scrolled

    def mainloop(
        self,
        interactive: bool = True,
        paste: curtsies.events.PasteEvent | None = None,
    ) -> None:
        if interactive:
            # Add custom help command
            # TODO: add methods to run the code
            self.initialize_interp()

            # run startup file
            self.process_event(events.RunStartupFileEvent())

        # handle paste
        if paste:
            self.process_event(paste)

        # do a display before waiting for first event
        self.process_event_and_paint(None)
        inputs = combined_events(self.input_generator)
        while self.module_gatherer.find_coroutine():
            e = inputs.send(0)
            if e is not None:
                self.process_event_and_paint(e)

        for e in inputs:
            self.process_event_and_paint(e)


def main(
    args: list[str] | None = None,
    locals_: dict[str, Any] | None = None,
    banner: str | None = None,
    welcome_message: str | None = None,
) -> Any:
    """
    banner is displayed directly after the version information.
    welcome_message is passed on to Repl and displayed in the statusbar.
    """
    translations.init()

    def curtsies_arguments(parser: argparse._ArgumentGroup) -> None:
        parser.add_argument(
            "--paste",
            "-p",
            action="store_true",
            help=_("start by pasting lines of a file into session"),
        )

    config, options, exec_args = bpargs.parse(
        args,
        (
            _("curtsies arguments"),
            _("Additional arguments specific to the curtsies-based REPL."),
            curtsies_arguments,
        ),
    )

    interp = None
    paste = None
    exit_value: tuple[Any, ...] = ()
    if exec_args:
        if not options:
            raise ValueError("don't pass in exec_args without options")
        if options.paste:
            paste = curtsies.events.PasteEvent()
            encoding = inspection.get_encoding_file(exec_args[0])
            with open(exec_args[0], encoding=encoding) as f:
                sourcecode = f.read()
            paste.events.extend(sourcecode)
        else:
            try:
                interp = Interp(locals=locals_)
                bpargs.exec_code(interp, exec_args)
            except SystemExit as e:
                exit_value = e.args
            if not options.interactive:
                return extract_exit_value(exit_value)
    else:
        # expected for interactive sessions (vanilla python does it)
        sys.path.insert(0, "")

    if not options.quiet:
        print(bpargs.version_banner())
    if banner is not None:
        print(banner)
    if welcome_message is None and not options.quiet and config.help_key:
        welcome_message = (
            _("Welcome to bpython!")
            + " "
            + _("Press <%s> for help.") % config.help_key
        )

    repl = FullCurtsiesRepl(config, locals_, welcome_message, interp)
    try:
        with repl.input_generator:
            with repl.window as win:
                with repl:
                    repl.height, repl.width = win.t.height, win.t.width
                    repl.mainloop(True, paste)
    except (SystemExitFromCodeRunner, SystemExit) as e:
        exit_value = e.args
    return extract_exit_value(exit_value)


def _combined_events(
    event_provider: SupportsEventGeneration, paste_threshold: int
) -> Generator[str | curtsies.events.Event | None, float | None, None]:
    """Combines consecutive keypress events into paste events."""
    timeout = yield "nonsense_event"  # so send can be used immediately
    queue: collections.deque = collections.deque()
    while True:
        e = event_provider.send(timeout)
        if isinstance(e, curtsies.events.Event):
            timeout = yield e
            continue
        elif e is None:
            timeout = yield None
            continue
        else:
            queue.append(e)
        e = event_provider.send(0)
        while not (e is None or isinstance(e, curtsies.events.Event)):
            queue.append(e)
            e = event_provider.send(0)
        if len(queue) >= paste_threshold:
            paste = curtsies.events.PasteEvent()
            paste.events.extend(queue)
            queue.clear()
            timeout = yield paste
        else:
            while len(queue):
                timeout = yield queue.popleft()


def combined_events(
    event_provider: SupportsEventGeneration, paste_threshold: int = 3
) -> SupportsEventGeneration:
    g = _combined_events(event_provider, paste_threshold)
    next(g)
    return g


if __name__ == "__main__":
    sys.exit(main())



import unittest
import unittest.mock
import os
from pathlib import Path

from bpython.translations import init


class FixLanguageTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        init(languages=["en"])


class MagicIterMock(unittest.mock.MagicMock):
    __next__ = unittest.mock.Mock(return_value=None)


TEST_CONFIG = Path(__file__).parent / "test.config"



import unittest

from collections import namedtuple
from bpython.curtsies import combined_events
from bpython.test import FixLanguageTestCase as TestCase

import curtsies.events


ScheduledEvent = namedtuple("ScheduledEvent", ["when", "event"])


class EventGenerator:
    def __init__(self, initial_events=(), scheduled_events=()):
        self._events = []
        self._current_tick = 0
        for e in initial_events:
            self.schedule_event(e, 0)
        for e, w in scheduled_events:
            self.schedule_event(e, w)

    def schedule_event(self, event, when):
        self._events.append(ScheduledEvent(when, event))
        self._events.sort()

    def send(self, timeout=None):
        if timeout not in [None, 0]:
            raise ValueError("timeout value %r not supported" % timeout)
        if not self._events:
            return None
        if self._events[0].when <= self._current_tick:
            return self._events.pop(0).event

        if timeout == 0:
            return None
        elif timeout is None:
            e = self._events.pop(0)
            self._current_tick = e.when
            return e.event
        else:
            raise ValueError("timeout value %r not supported" % timeout)

    def tick(self, dt=1):
        self._current_tick += dt
        return self._current_tick


class TestCurtsiesPasteDetection(TestCase):
    def test_paste_threshold(self):
        eg = EventGenerator(list("abc"))
        cb = combined_events(eg, paste_threshold=3)
        e = next(cb)
        self.assertIsInstance(e, curtsies.events.PasteEvent)
        self.assertEqual(e.events, list("abc"))
        self.assertEqual(next(cb), None)

        eg = EventGenerator(list("abc"))
        cb = combined_events(eg, paste_threshold=4)
        self.assertEqual(next(cb), "a")
        self.assertEqual(next(cb), "b")
        self.assertEqual(next(cb), "c")
        self.assertEqual(next(cb), None)

    def test_set_timeout(self):
        eg = EventGenerator("a", zip("bcdefg", [1, 2, 3, 3, 3, 4]))
        eg.schedule_event(curtsies.events.SigIntEvent(), 5)
        eg.schedule_event("h", 6)
        cb = combined_events(eg, paste_threshold=3)
        self.assertEqual(next(cb), "a")
        self.assertEqual(cb.send(0), None)
        self.assertEqual(next(cb), "b")
        self.assertEqual(cb.send(0), None)
        eg.tick()
        self.assertEqual(cb.send(0), "c")
        self.assertEqual(cb.send(0), None)
        eg.tick()
        self.assertIsInstance(cb.send(0), curtsies.events.PasteEvent)
        self.assertEqual(cb.send(0), None)
        self.assertEqual(cb.send(None), "g")
        self.assertEqual(cb.send(0), None)
        eg.tick(1)
        self.assertIsInstance(cb.send(0), curtsies.events.SigIntEvent)
        self.assertEqual(cb.send(0), None)
        self.assertEqual(cb.send(None), "h")
        self.assertEqual(cb.send(None), None)


if __name__ == "__main__":
    unittest.main()



# The MIT License
#
# Copyright (c) 2015 the bpython authors.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""simple evaluation of side-effect free code

In order to provide fancy completion, some code can be executed safely.
"""

import ast
import builtins
from typing import Any

from . import line as line_properties
from .inspection import getattr_safe

_numeric_types = (int, float, complex)


class EvaluationError(Exception):
    """Raised if an exception occurred in safe_eval."""


def safe_eval(expr: str, namespace: dict[str, Any]) -> Any:
    """Not all that safe, just catches some errors"""
    try:
        return eval(expr, namespace)
    except (NameError, AttributeError, SyntaxError):
        # If debugging safe_eval, raise this!
        # raise
        raise EvaluationError


# This function is under the Python License, Version 2
# This license requires modifications to the code be reported.
# Based on ast.literal_eval
# Modifications:
# * checks that objects used as operands of + and - are numbers
#   instead of checking they are constructed with number literals
# * new docstring describing different functionality
# * looks up names from namespace
# * indexing syntax is allowed
# * evaluates tuple() and list()
def simple_eval(node_or_string, namespace=None):
    """
    Safely evaluate an expression node or a string containing a Python
    expression without triggering any user code.

    The string or node provided may only consist of:
    * the following Python literal structures: strings, numbers, tuples,
        lists, dicts, and sets
    * variable names causing lookups in the passed in namespace or builtins
    * getitem calls using the [] syntax on objects of the types above

    Like Python 3's literal_eval, unary and binary + and - operations are
    allowed on all builtin numeric types.

    The optional namespace dict-like ought not to cause side effects on lookup.
    """
    if namespace is None:
        namespace = {}
    if isinstance(node_or_string, str):
        node_or_string = ast.parse(node_or_string, mode="eval")
    if isinstance(node_or_string, ast.Expression):
        node_or_string = node_or_string.body

    def _convert(node):
        if isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, ast.List):
            return list(map(_convert, node.elts))
        elif isinstance(node, ast.Dict):
            return {
                _convert(k): _convert(v) for k, v in zip(node.keys, node.values)
            }
        elif isinstance(node, ast.Set):
            return set(map(_convert, node.elts))
        elif (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "set"
            and node.args == node.keywords == []
        ):
            return set()

        # this is a deviation from literal_eval: we evaluate tuple() and list()
        elif (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "tuple"
            and node.args == node.keywords == []
        ):
            return tuple()
        elif (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "list"
            and node.args == node.keywords == []
        ):
            return list()

        # this is a deviation from literal_eval: we allow non-literals
        elif isinstance(node, ast.Name):
            try:
                return namespace[node.id]
            except KeyError:
                try:
                    return getattr(builtins, node.id)
                except AttributeError:
                    raise EvaluationError("can't lookup %s" % node.id)

        # unary + and - are allowed on any type
        elif isinstance(node, ast.UnaryOp) and isinstance(
            node.op, (ast.UAdd, ast.USub)
        ):
            # ast.literal_eval does ast typechecks here, we use type checks
            operand = _convert(node.operand)
            if not type(operand) in _numeric_types:
                raise ValueError("unary + and - only allowed on builtin nums")
            if isinstance(node.op, ast.UAdd):
                return +operand
            else:
                return -operand
        elif isinstance(node, ast.BinOp) and isinstance(
            node.op, (ast.Add, ast.Sub)
        ):
            # this is a deviation from literal_eval: ast.literal_eval accepts
            # (+/-) int, float and complex literals as left operand, and complex
            # as right operation, we evaluate as much as possible
            left = _convert(node.left)
            right = _convert(node.right)
            if not (
                isinstance(left, _numeric_types)
                and isinstance(right, _numeric_types)
            ):
                raise ValueError("binary + and - only allowed on builtin nums")
            if isinstance(node.op, ast.Add):
                return left + right
            else:
                return left - right

        # this is a deviation from literal_eval: we allow indexing
        elif isinstance(node, ast.Subscript) and isinstance(
            node.slice, (ast.Constant, ast.Name)
        ):
            obj = _convert(node.value)
            index = _convert(node.slice)
            return safe_getitem(obj, index)

        # this is a deviation from literal_eval: we allow attribute access
        if isinstance(node, ast.Attribute):
            obj = _convert(node.value)
            attr = node.attr
            return getattr_safe(obj, attr)

        raise ValueError(f"malformed node or string: {node!r}")

    return _convert(node_or_string)


def safe_getitem(obj, index):
    """Safely tries to access obj[index]"""
    if type(obj) in (list, tuple, dict, bytes, str):
        try:
            return obj[index]
        except (KeyError, IndexError):
            raise EvaluationError(f"can't lookup key {index!r} on {obj!r}")
    raise ValueError(f"unsafe to lookup on object of type {type(obj)}")


def find_attribute_with_name(node, name):
    if isinstance(node, ast.Attribute) and node.attr == name:
        return node
    for item in ast.iter_child_nodes(node):
        r = find_attribute_with_name(item, name)
        if r:
            return r


def evaluate_current_expression(
    cursor_offset: int, line: str, namespace: dict[str, Any] | None = None
) -> Any:
    """
    Return evaluated expression to the right of the dot of current attribute.

    Only evaluates builtin objects, and do any attribute lookup.
    """
    # Builds asts from with increasing numbers of characters back from cursor.
    # Find the biggest valid ast.
    # Once our attribute access is found, return its .value subtree

    # in case attribute is blank, e.g. foo.| -> foo.xxx|
    temp_line = line[:cursor_offset] + "xxx" + line[cursor_offset:]
    temp_cursor = cursor_offset + 3
    temp_attribute = line_properties.current_expression_attribute(
        temp_cursor, temp_line
    )
    if temp_attribute is None:
        raise EvaluationError("No current attribute")
    attr_before_cursor = temp_line[temp_attribute.start : temp_cursor]

    def parse_trees(cursor_offset, line):
        for i in range(cursor_offset - 1, -1, -1):
            try:
                tree = ast.parse(line[i:cursor_offset])
                yield tree
            except SyntaxError:
                continue

    largest_ast = None
    for tree in parse_trees(temp_cursor, temp_line):
        attribute_access = find_attribute_with_name(tree, attr_before_cursor)
        if attribute_access:
            largest_ast = attribute_access.value

    if largest_ast is None:
        raise EvaluationError(
            "Corresponding ASTs to right of cursor are invalid"
        )
    try:
        return simple_eval(largest_ast, namespace)
    except ValueError:
        raise EvaluationError("Could not safely evaluate")


def evaluate_current_attribute(cursor_offset, line, namespace=None):
    """Safely evaluates the expression having an attributed accessed"""
    # this function runs user code in case of custom descriptors,
    # so could fail in any way

    obj = evaluate_current_expression(cursor_offset, line, namespace)
    attr = line_properties.current_expression_attribute(cursor_offset, line)
    if attr is None:
        raise EvaluationError("No attribute found to look up")
    try:
        return getattr(obj, attr.word)
    except AttributeError:
        raise EvaluationError(f"can't lookup attribute {attr.word} on {obj!r}")



import os
import tempfile
import unittest

from pathlib import Path
from bpython.importcompletion import ModuleGatherer


class TestSimpleComplete(unittest.TestCase):
    def setUp(self):
        self.module_gatherer = ModuleGatherer()
        self.module_gatherer.modules = [
            "zzabc",
            "zzabd",
            "zzefg",
            "zzabc.e",
            "zzabc.f",
            "zzefg.a1",
            "zzefg.a2",
        ]

    def test_simple_completion(self):
        self.assertSetEqual(
            self.module_gatherer.complete(10, "import zza"), {"zzabc", "zzabd"}
        )
        self.assertSetEqual(
            self.module_gatherer.complete(11, "import  zza"), {"zzabc", "zzabd"}
        )

    def test_import_empty(self):
        self.assertSetEqual(
            self.module_gatherer.complete(13, "import zzabc."),
            {"zzabc.e", "zzabc.f"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(14, "import  zzabc."),
            {"zzabc.e", "zzabc.f"},
        )

    def test_import(self):
        self.assertSetEqual(
            self.module_gatherer.complete(14, "import zzefg.a"),
            {"zzefg.a1", "zzefg.a2"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(15, "import  zzefg.a"),
            {"zzefg.a1", "zzefg.a2"},
        )

    @unittest.expectedFailure
    def test_import_blank(self):
        self.assertSetEqual(
            self.module_gatherer.complete(7, "import "),
            {"zzabc", "zzabd", "zzefg"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(8, "import  "),
            {"zzabc", "zzabd", "zzefg"},
        )

    @unittest.expectedFailure
    def test_from_import_empty(self):
        self.assertSetEqual(
            self.module_gatherer.complete(5, "from "),
            {"zzabc", "zzabd", "zzefg"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(6, "from  "),
            {"zzabc", "zzabd", "zzefg"},
        )

    @unittest.expectedFailure
    def test_from_module_import_empty(self):
        self.assertSetEqual(
            self.module_gatherer.complete(18, "from zzabc import "), {"e", "f"}
        )
        self.assertSetEqual(
            self.module_gatherer.complete(19, "from  zzabc import "), {"e", "f"}
        )
        self.assertSetEqual(
            self.module_gatherer.complete(19, "from zzabc  import "), {"e", "f"}
        )
        self.assertSetEqual(
            self.module_gatherer.complete(19, "from zzabc import  "), {"e", "f"}
        )

    def test_from_module_import(self):
        self.assertSetEqual(
            self.module_gatherer.complete(19, "from zzefg import a"),
            {"a1", "a2"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(20, "from  zzefg import a"),
            {"a1", "a2"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(20, "from zzefg  import a"),
            {"a1", "a2"},
        )
        self.assertSetEqual(
            self.module_gatherer.complete(20, "from zzefg import  a"),
            {"a1", "a2"},
        )


class TestRealComplete(unittest.TestCase):
    def setUp(self):
        self.module_gatherer = ModuleGatherer()
        while self.module_gatherer.find_coroutine():
            pass
        __import__("sys")
        __import__("os")

    def test_from_attribute(self):
        self.assertSetEqual(
            self.module_gatherer.complete(19, "from sys import arg"), {"argv"}
        )

    def test_from_attr_module(self):
        self.assertSetEqual(
            self.module_gatherer.complete(9, "from os.p"), {"os.path"}
        )

    def test_from_package(self):
        self.assertSetEqual(
            self.module_gatherer.complete(17, "from xml import d"), {"dom"}
        )


class TestAvoidSymbolicLinks(unittest.TestCase):
    def setUp(self):
        with tempfile.TemporaryDirectory() as import_test_folder:
            base_path = Path(import_test_folder)
            (base_path / "Level0" / "Level1" / "Level2").mkdir(parents=True)
            (base_path / "Left").mkdir(parents=True)
            (base_path / "Right").mkdir(parents=True)

            current_path = base_path / "Level0"
            (current_path / "__init__.py").touch()

            current_path = current_path / "Level1"
            (current_path / "__init__.py").touch()

            current_path = current_path / "Level2"
            (current_path / "__init__.py").touch()
            # Level0/Level1/Level2/Level3 -> Level0/Level1
            (current_path / "Level3").symlink_to(
                base_path / "Level0" / "Level1", target_is_directory=True
            )

            current_path = base_path / "Right"
            (current_path / "__init__.py").touch()
            # Right/toLeft -> Left
            (current_path / "toLeft").symlink_to(
                base_path / "Left", target_is_directory=True
            )

            current_path = base_path / "Left"
            (current_path / "__init__.py").touch()
            # Left/toRight -> Right
            (current_path / "toRight").symlink_to(
                base_path / "Right", target_is_directory=True
            )

            self.module_gatherer = ModuleGatherer((base_path.absolute(),))
            while self.module_gatherer.find_coroutine():
                pass

    def test_simple_symbolic_link_loop(self):
        filepaths = [
            "Left.toRight.toLeft",
            "Left.toRight",
            "Left",
            "Level0.Level1.Level2.Level3",
            "Level0.Level1.Level2",
            "Level0.Level1",
            "Level0",
            "Right",
            "Right.toLeft",
            "Right.toLeft.toRight",
        ]

        for thing in self.module_gatherer.modules:
            self.assertIn(thing, filepaths)
            if thing == "Left.toRight.toLeft":
                filepaths.remove("Right.toLeft")
                filepaths.remove("Right.toLeft.toRight")
            if thing == "Right.toLeft.toRight":
                filepaths.remove("Left.toRight.toLeft")
                filepaths.remove("Left.toRight")
            filepaths.remove(thing)
        self.assertFalse(filepaths)


class TestBugReports(unittest.TestCase):
    def test_issue_847(self):
        with tempfile.TemporaryDirectory() as import_test_folder:
            #   ./xyzzy
            #   ./xyzzy/__init__.py
            #   ./xyzzy/plugh
            #   ./xyzzy/plugh/__init__.py
            #   ./xyzzy/plugh/bar.py
            #   ./xyzzy/plugh/foo.py

            base_path = Path(import_test_folder)
            (base_path / "xyzzy" / "plugh").mkdir(parents=True)
            (base_path / "xyzzy" / "__init__.py").touch()
            (base_path / "xyzzy" / "plugh" / "__init__.py").touch()
            (base_path / "xyzzy" / "plugh" / "bar.py").touch()
            (base_path / "xyzzy" / "plugh" / "foo.py").touch()

            module_gatherer = ModuleGatherer((base_path.absolute(),))
            while module_gatherer.find_coroutine():
                pass

            self.assertSetEqual(
                module_gatherer.complete(17, "from xyzzy.plugh."),
                {"xyzzy.plugh.bar", "xyzzy.plugh.foo"},
            )


if __name__ == "__main__":
    unittest.main()



# The MIT License
#
# Copyright (c) 2008 Bob Farrell
# Copyright (c) 2012-2021 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# To gradually migrate to mypy we aren't setting these globally yet
# mypy: disallow_untyped_defs=True
# mypy: disallow_untyped_calls=True

"""
Module to handle command line argument parsing, for all front-ends.
"""

import argparse
import code
import importlib.util
import logging
import os
import sys
from pathlib import Path
from collections.abc import Callable
from types import ModuleType

from . import __version__, __copyright__
from .config import default_config_path, Config
from .translations import _
from ._typing_compat import Never

logger = logging.getLogger(__name__)


class ArgumentParserFailed(ValueError):
    """Raised by the RaisingOptionParser for a bogus commandline."""


class RaisingArgumentParser(argparse.ArgumentParser):
    def error(self, msg: str) -> Never:
        raise ArgumentParserFailed()


def version_banner(base: str = "bpython") -> str:
    return _("{} version {} on top of Python {} {}").format(
        base,
        __version__,
        sys.version.split()[0],
        sys.executable,
    )


def copyright_banner() -> str:
    return _("{} See AUTHORS.rst for details.").format(__copyright__)


def log_version(module: ModuleType, name: str) -> None:
    logger.info("%s: %s", name, module.__version__ if hasattr(module, "__version__") else "unknown version")  # type: ignore


Options = tuple[str, str, Callable[[argparse._ArgumentGroup], None]]


def parse(
    args: list[str] | None,
    extras: Options | None = None,
    ignore_stdin: bool = False,
) -> tuple[Config, argparse.Namespace, list[str]]:
    """Receive an argument list - if None, use sys.argv - parse all args and
    take appropriate action. Also receive optional extra argument: this should
    be a tuple of (title, description, callback)
        title:          The title for the argument group
        description:    A full description of the argument group
        callback:       A callback that adds argument to the argument group

    e.g.:

    def callback(group):
        group.add_argument('-f', action='store_true', dest='f', help='Explode')
        group.add_argument('-l', action='store_true', dest='l', help='Love')

    parse(
        ['-i', '-m', 'foo.py'],
        (
            'Front end-specific options',
            'A full description of what these options are for',
            callback
        ),
    )


    Return a tuple of (config, options, exec_args) wherein "config" is the
    config object either parsed from a default/specified config file or default
    config options, "options" is the parsed options from
    ArgumentParser.parse_args, and "exec_args" are the args (if any) to be parsed
    to the executed file (if any).
    """
    if args is None:
        args = sys.argv[1:]

    parser = RaisingArgumentParser(
        usage=_(
            "Usage: %(prog)s [options] [file [args]]\n"
            "NOTE: If bpython sees an argument it does "
            "not know, execution falls back to the "
            "regular Python interpreter."
        )
    )
    parser.add_argument(
        "--config",
        default=default_config_path(),
        type=Path,
        help=_("Use CONFIG instead of default config file."),
    )
    parser.add_argument(
        "--interactive",
        "-i",
        action="store_true",
        help=_("Drop to bpython shell after running file instead of exiting."),
    )
    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help=_("Don't print version banner."),
    )
    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        help=_("Print version and exit."),
    )
    parser.add_argument(
        "--log-level",
        "-l",
        choices=("debug", "info", "warning", "error", "critical"),
        default="error",
        help=_("Set log level for logging"),
    )
    parser.add_argument(
        "--log-output",
        "-L",
        help=_("Log output file"),
    )

    if extras is not None:
        extras_group = parser.add_argument_group(extras[0], extras[1])
        extras[2](extras_group)

    # collect all the remaining arguments into a list
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        help=_(
            "File to execute and additional arguments passed on to the executed script."
        ),
    )

    try:
        options = parser.parse_args(args)
    except ArgumentParserFailed:
        # Just let Python handle this
        os.execv(sys.executable, [sys.executable] + args)

    if options.version:
        print(version_banner())
        print(copyright_banner())
        raise SystemExit

    if not ignore_stdin and not (sys.stdin.isatty() and sys.stdout.isatty()):
        # Just let Python handle this
        os.execv(sys.executable, [sys.executable] + args)

    # Configure logging handler
    bpython_logger = logging.getLogger("bpython")
    curtsies_logger = logging.getLogger("curtsies")
    bpython_logger.setLevel(options.log_level.upper())
    curtsies_logger.setLevel(options.log_level.upper())
    if options.log_output:
        handler = logging.FileHandler(filename=options.log_output)
        handler.setFormatter(
            logging.Formatter(
                "%(asctime)s: %(name)s: %(levelname)s: %(message)s"
            )
        )
        bpython_logger.addHandler(handler)
        curtsies_logger.addHandler(handler)
        bpython_logger.propagate = curtsies_logger.propagate = False
    else:
        bpython_logger.addHandler(logging.NullHandler())
        curtsies_logger.addHandler(logging.NullHandler())

    import cwcwidth
    import greenlet
    import pygments
    import requests
    import xdg

    logger.info("Starting bpython %s", __version__)
    logger.info("Python %s: %s", sys.executable, sys.version_info)
    # versions of required dependencies
    try:
        import curtsies

        log_version(curtsies, "curtsies")
    except ImportError:
        # may happen on Windows
        logger.info("curtsies: not available")
    log_version(cwcwidth, "cwcwidth")
    log_version(greenlet, "greenlet")
    log_version(pygments, "pygments")
    log_version(xdg, "pyxdg")
    log_version(requests, "requests")

    # versions of optional dependencies
    try:
        import pyperclip

        log_version(pyperclip, "pyperclip")
    except ImportError:
        logger.info("pyperclip: not available")
    try:
        import jedi

        log_version(jedi, "jedi")
    except ImportError:
        logger.info("jedi: not available")
    try:
        import watchdog

        logger.info("watchdog: available")
    except ImportError:
        logger.info("watchdog: not available")

    logger.info("environment:")
    for key, value in sorted(os.environ.items()):
        if key.startswith("LC") or key.startswith("LANG") or key == "TERM":
            logger.info("%s: %s", key, value)

    return Config(options.config), options, options.args


def exec_code(
    interpreter: code.InteractiveInterpreter, args: list[str]
) -> None:
    """
    Helper to execute code in a given interpreter, e.g. to implement the behavior of python3 [-i] file.py

    args should be a [faked] sys.argv.
    """
    try:
        with open(args[0]) as sourcefile:
            source = sourcefile.read()
    except OSError as e:
        # print an error and exit (if -i is specified the calling code will continue)
        print(f"bpython: can't open file '{args[0]}: {e}", file=sys.stderr)
        raise SystemExit(e.errno)
    old_argv, sys.argv = sys.argv, args
    sys.path.insert(0, os.path.abspath(os.path.dirname(args[0])))
    spec = importlib.util.spec_from_loader("__main__", loader=None)
    assert spec
    mod = importlib.util.module_from_spec(spec)
    sys.modules["__main__"] = mod
    interpreter.locals.update(mod.__dict__)  # type: ignore  # TODO use a more specific type that has a .locals attribute
    interpreter.locals["__file__"] = args[0]  # type: ignore  # TODO use a more specific type that has a .locals attribute
    interpreter.runsource(source, args[0], "exec")
    sys.argv = old_argv



import inspect
import os
import sys
import unittest
from collections.abc import Sequence
from typing import List

from bpython import inspection
from bpython.test.fodder import encoding_ascii
from bpython.test.fodder import encoding_latin1
from bpython.test.fodder import encoding_utf8

pypy = "PyPy" in sys.version
_is_py311 = sys.version_info[:2] >= (3, 11)

try:
    import numpy
except ImportError:
    numpy = None  # type: ignore


foo_ascii_only = '''def foo():
    """Test"""
    pass
'''

foo_non_ascii = '''def foo():
    """Test äöü"""
    pass
'''


class Callable:
    def __call__(self):
        pass


class Noncallable:
    pass


def spam():
    pass


class CallableMethod:
    def method(self):
        pass


class TestInspection(unittest.TestCase):
    def test_parsekeywordpairs(self):
        # See issue #109
        def fails(spam=["-a", "-b"]):
            pass

        argspec = inspection.getfuncprops("fails", fails)
        self.assertIsNotNone(argspec)
        defaults = argspec.argspec.defaults
        self.assertEqual(str(defaults[0]), '["-a", "-b"]')

    def test_pasekeywordpairs_string(self):
        def spam(eggs="foo, bar"):
            pass

        defaults = inspection.getfuncprops("spam", spam).argspec.defaults
        self.assertEqual(repr(defaults[0]), '"foo, bar"')

    def test_parsekeywordpairs_multiple_keywords(self):
        def spam(eggs=23, foobar="yay"):
            pass

        defaults = inspection.getfuncprops("spam", spam).argspec.defaults
        self.assertEqual(repr(defaults[0]), "23")
        self.assertEqual(repr(defaults[1]), '"yay"')

    def test_pasekeywordpairs_annotation(self):
        def spam(eggs: str = "foo, bar"):
            pass

        defaults = inspection.getfuncprops("spam", spam).argspec.defaults
        self.assertEqual(repr(defaults[0]), '"foo, bar"')

    def test_get_encoding_ascii(self):
        self.assertEqual(inspection.get_encoding(encoding_ascii), "ascii")
        self.assertEqual(inspection.get_encoding(encoding_ascii.foo), "ascii")

    def test_get_encoding_latin1(self):
        self.assertEqual(inspection.get_encoding(encoding_latin1), "latin1")
        self.assertEqual(inspection.get_encoding(encoding_latin1.foo), "latin1")

    def test_get_encoding_utf8(self):
        self.assertEqual(inspection.get_encoding(encoding_utf8), "utf-8")
        self.assertEqual(inspection.get_encoding(encoding_utf8.foo), "utf-8")

    def test_get_source_ascii(self):
        self.assertEqual(inspect.getsource(encoding_ascii.foo), foo_ascii_only)

    def test_get_source_utf8(self):
        self.assertEqual(inspect.getsource(encoding_utf8.foo), foo_non_ascii)

    def test_get_source_latin1(self):
        self.assertEqual(inspect.getsource(encoding_latin1.foo), foo_non_ascii)

    def test_get_source_file(self):
        path = os.path.join(os.path.dirname(__file__), "fodder")

        encoding = inspection.get_encoding_file(
            os.path.join(path, "encoding_ascii.py")
        )
        self.assertEqual(encoding, "ascii")
        encoding = inspection.get_encoding_file(
            os.path.join(path, "encoding_latin1.py")
        )
        self.assertEqual(encoding, "latin1")
        encoding = inspection.get_encoding_file(
            os.path.join(path, "encoding_utf8.py")
        )
        self.assertEqual(encoding, "utf-8")

    @unittest.skipIf(pypy, "pypy builtin signatures aren't complete")
    def test_getfuncprops_print(self):
        props = inspection.getfuncprops("print", print)

        self.assertEqual(props.func, "print")
        self.assertIn("end", props.argspec.kwonly)
        self.assertIn("file", props.argspec.kwonly)
        self.assertIn("flush", props.argspec.kwonly)
        self.assertIn("sep", props.argspec.kwonly)
        if _is_py311:
            self.assertEqual(
                repr(props.argspec.kwonly_defaults["file"]), "None"
            )
        else:
            self.assertEqual(
                repr(props.argspec.kwonly_defaults["file"]), "sys.stdout"
            )
        self.assertEqual(repr(props.argspec.kwonly_defaults["flush"]), "False")

    @unittest.skipUnless(
        numpy is not None and numpy.__version__ >= "1.18",
        "requires numpy >= 1.18",
    )
    def test_getfuncprops_numpy_array(self):
        props = inspection.getfuncprops("array", numpy.array)

        self.assertEqual(props.func, "array")
        # This check might need an update in the future, but at least numpy >= 1.18 has
        # np.array(object, dtype=None, *, ...).
        self.assertEqual(props.argspec.args, ["object", "dtype"])

    def test_issue_966_freestanding(self):
        def fun(number, lst=[]):
            """
            Return a list of numbers

            Example:
            ========
            C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
            """
            return lst + [number]

        def fun_annotations(number: int, lst: list[int] = []) -> list[int]:
            """
            Return a list of numbers

            Example:
            ========
            C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
            """
            return lst + [number]

        props = inspection.getfuncprops("fun", fun)
        self.assertEqual(props.func, "fun")
        self.assertEqual(props.argspec.args, ["number", "lst"])
        self.assertEqual(repr(props.argspec.defaults[0]), "[]")

        props = inspection.getfuncprops("fun_annotations", fun_annotations)
        self.assertEqual(props.func, "fun_annotations")
        self.assertEqual(props.argspec.args, ["number", "lst"])
        self.assertEqual(repr(props.argspec.defaults[0]), "[]")

    def test_issue_966_class_method(self):
        class Issue966(Sequence):
            @classmethod
            def cmethod(cls, number: int, lst: list[int] = []):
                """
                Return a list of numbers

                Example:
                ========
                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
                """
                return lst + [number]

            @classmethod
            def bmethod(cls, number, lst):
                """
                Return a list of numbers

                Example:
                ========
                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
                """
                return lst + [number]

        props = inspection.getfuncprops(
            "bmethod", inspection.getattr_safe(Issue966, "bmethod")
        )
        self.assertEqual(props.func, "bmethod")
        self.assertEqual(props.argspec.args, ["number", "lst"])

        props = inspection.getfuncprops(
            "cmethod", inspection.getattr_safe(Issue966, "cmethod")
        )
        self.assertEqual(props.func, "cmethod")
        self.assertEqual(props.argspec.args, ["number", "lst"])
        self.assertEqual(repr(props.argspec.defaults[0]), "[]")

    def test_issue_966_static_method(self):
        class Issue966(Sequence):
            @staticmethod
            def cmethod(number: int, lst: list[int] = []):
                """
                Return a list of numbers

                Example:
                ========
                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
                """
                return lst + [number]

            @staticmethod
            def bmethod(number, lst):
                """
                Return a list of numbers

                Example:
                ========
                C.cmethod(1337, [1, 2]) # => [1, 2, 1337]
                """
                return lst + [number]

        props = inspection.getfuncprops(
            "bmethod", inspection.getattr_safe(Issue966, "bmethod")
        )
        self.assertEqual(props.func, "bmethod")
        self.assertEqual(props.argspec.args, ["number", "lst"])

        props = inspection.getfuncprops(
            "cmethod", inspection.getattr_safe(Issue966, "cmethod")
        )
        self.assertEqual(props.func, "cmethod")
        self.assertEqual(props.argspec.args, ["number", "lst"])
        self.assertEqual(repr(props.argspec.defaults[0]), "[]")


class A:
    a = "a"


class B(A):
    b = "b"


class Property:
    @property
    def prop(self):
        raise AssertionError("Property __get__ executed")


class Slots:
    __slots__ = ["s1", "s2", "s3"]


class SlotsSubclass(Slots):
    @property
    def s4(self):
        raise AssertionError("Property __get__ executed")


class OverriddenGetattr:
    def __getattr__(self, attr):
        raise AssertionError("custom __getattr__ executed")

    a = 1


class OverriddenGetattribute:
    def __getattribute__(self, attr):
        raise AssertionError("custom __getattribute__ executed")

    a = 1


class OverriddenMRO:
    def __mro__(self):
        raise AssertionError("custom mro executed")

    a = 1


member_descriptor = type(Slots.s1)  # type: ignore


class TestSafeGetAttribute(unittest.TestCase):
    def test_lookup_on_object(self):
        a = A()
        a.x = 1
        self.assertEqual(inspection.getattr_safe(a, "x"), 1)
        self.assertEqual(inspection.getattr_safe(a, "a"), "a")
        b = B()
        b.y = 2
        self.assertEqual(inspection.getattr_safe(b, "y"), 2)
        self.assertEqual(inspection.getattr_safe(b, "a"), "a")
        self.assertEqual(inspection.getattr_safe(b, "b"), "b")

        self.assertEqual(inspection.hasattr_safe(b, "y"), True)
        self.assertEqual(inspection.hasattr_safe(b, "b"), True)

    def test_avoid_running_properties(self):
        p = Property()
        self.assertEqual(inspection.getattr_safe(p, "prop"), Property.prop)
        self.assertEqual(inspection.hasattr_safe(p, "prop"), True)

    def test_lookup_with_slots(self):
        s = Slots()
        s.s1 = "s1"
        self.assertEqual(inspection.getattr_safe(s, "s1"), "s1")
        with self.assertRaises(AttributeError):
            inspection.getattr_safe(s, "s2")

        self.assertEqual(inspection.hasattr_safe(s, "s1"), True)
        self.assertEqual(inspection.hasattr_safe(s, "s2"), False)

    def test_lookup_on_slots_classes(self):
        sga = inspection.getattr_safe
        s = SlotsSubclass()
        self.assertIsInstance(sga(Slots, "s1"), member_descriptor)
        self.assertIsInstance(sga(SlotsSubclass, "s1"), member_descriptor)
        self.assertIsInstance(sga(SlotsSubclass, "s4"), property)
        self.assertIsInstance(sga(s, "s4"), property)

        self.assertEqual(inspection.hasattr_safe(s, "s1"), False)
        self.assertEqual(inspection.hasattr_safe(s, "s4"), True)

    def test_lookup_on_overridden_methods(self):
        sga = inspection.getattr_safe
        self.assertEqual(sga(OverriddenGetattr(), "a"), 1)
        self.assertEqual(sga(OverriddenGetattribute(), "a"), 1)
        self.assertEqual(sga(OverriddenMRO(), "a"), 1)
        with self.assertRaises(AttributeError):
            sga(OverriddenGetattr(), "b")
        with self.assertRaises(AttributeError):
            sga(OverriddenGetattribute(), "b")
        with self.assertRaises(AttributeError):
            sga(OverriddenMRO(), "b")

        self.assertEqual(
            inspection.hasattr_safe(OverriddenGetattr(), "b"), False
        )
        self.assertEqual(
            inspection.hasattr_safe(OverriddenGetattribute(), "b"), False
        )
        self.assertEqual(inspection.hasattr_safe(OverriddenMRO(), "b"), False)


if __name__ == "__main__":
    unittest.main()



# The MIT License
#
# Copyright (c) 2009-2011 the bpython authors.
# Copyright (c) 2015 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import inspect
import keyword
import pydoc
import re
from dataclasses import dataclass
from typing import (
    Any,
    ContextManager,
    Literal,
)
from collections.abc import Callable
from types import MemberDescriptorType, TracebackType

from pygments.token import Token
from pygments.lexers import Python3Lexer

from .lazyre import LazyReCompile


class _Repr:
    """
    Helper for `ArgSpec`: Returns the given value in `__repr__()`.
    """

    __slots__ = ("value",)

    def __init__(self, value: str) -> None:
        self.value = value

    def __repr__(self) -> str:
        return self.value

    __str__ = __repr__


@dataclass
class ArgSpec:
    args: list[str]
    varargs: str | None
    varkwargs: str | None
    defaults: list[_Repr] | None
    kwonly: list[str]
    kwonly_defaults: dict[str, _Repr] | None
    annotations: dict[str, Any] | None


@dataclass
class FuncProps:
    func: str
    argspec: ArgSpec
    is_bound_method: bool


class AttrCleaner(ContextManager[None]):
    """A context manager that tries to make an object not exhibit side-effects
    on attribute lookup.

    Unless explicitly required, prefer `getattr_safe`."""

    def __init__(self, obj: Any) -> None:
        self._obj = obj

    def __enter__(self) -> None:
        """Try to make an object not exhibit side-effects on attribute
        lookup."""
        type_ = type(self._obj)
        # Dark magic:
        # If __getattribute__ doesn't exist on the class and __getattr__ does
        # then __getattr__ will be called when doing
        #   getattr(type_, '__getattribute__', None)
        # so we need to first remove the __getattr__, then the
        # __getattribute__, then look up the attributes and then restore the
        # original methods. :-(
        # The upshot being that introspecting on an object to display its
        # attributes will avoid unwanted side-effects.
        __getattr__ = getattr(type_, "__getattr__", None)
        if __getattr__ is not None:
            try:
                setattr(type_, "__getattr__", (lambda *_, **__: None))
            except (TypeError, AttributeError):
                __getattr__ = None
        __getattribute__ = getattr(type_, "__getattribute__", None)
        if __getattribute__ is not None:
            try:
                setattr(type_, "__getattribute__", object.__getattribute__)
            except (TypeError, AttributeError):
                # XXX: This happens for e.g. built-in types
                __getattribute__ = None
        self._attribs = (__getattribute__, __getattr__)
        # /Dark magic

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Restore an object's magic methods."""
        type_ = type(self._obj)
        __getattribute__, __getattr__ = self._attribs
        # Dark magic:
        if __getattribute__ is not None:
            setattr(type_, "__getattribute__", __getattribute__)
        if __getattr__ is not None:
            setattr(type_, "__getattr__", __getattr__)
        # /Dark magic
        return False


def parsekeywordpairs(signature: str) -> dict[str, str]:
    preamble = True
    stack = []
    substack: list[str] = []
    parendepth = 0
    annotation = False
    for token, value in Python3Lexer().get_tokens(signature):
        if preamble:
            if token is Token.Punctuation and value == "(":
                # First "(" starts the list of arguments
                preamble = False
            continue

        if token is Token.Punctuation:
            if value in "({[":
                parendepth += 1
            elif value in ")}]":
                parendepth -= 1
            elif value == ":":
                if parendepth == -1:
                    # End of signature reached
                    break
                elif parendepth == 0:
                    # Start of type annotation
                    annotation = True

            if (value, parendepth) in ((",", 0), (")", -1)):
                # End of current argument
                stack.append(substack)
                substack = []
                # If type annotation didn't end before, it does now.
                annotation = False
                continue
        elif token is Token.Operator and value == "=" and parendepth == 0:
            # End of type annotation
            annotation = False

        if value and not annotation and (parendepth > 0 or value.strip()):
            substack.append(value)

    return {item[0]: "".join(item[2:]) for item in stack if len(item) >= 3}


def _fix_default_values(f: Callable, argspec: ArgSpec) -> ArgSpec:
    """Functions taking default arguments that are references to other objects
    will cause breakage, so we swap out the object itself with the name it was
    referenced with in the source by parsing the source itself!"""

    if argspec.defaults is None and argspec.kwonly_defaults is None:
        # No keyword args, no need to do anything
        return argspec

    try:
        src, _ = inspect.getsourcelines(f)
    except (OSError, IndexError):
        # IndexError is raised in inspect.findsource(), can happen in
        # some situations. See issue #94.
        return argspec
    except TypeError:
        # No source code is available, so replace the default values with what we have.
        if argspec.defaults is not None:
            argspec.defaults = [_Repr(str(value)) for value in argspec.defaults]
        if argspec.kwonly_defaults is not None:
            argspec.kwonly_defaults = {
                key: _Repr(str(value))
                for key, value in argspec.kwonly_defaults.items()
            }
        return argspec

    kwparsed = parsekeywordpairs("".join(src))

    if argspec.defaults is not None:
        values = list(argspec.defaults)
        keys = argspec.args[-len(values) :]
        for i, key in enumerate(keys):
            values[i] = _Repr(kwparsed[key])

        argspec.defaults = values
    if argspec.kwonly_defaults is not None:
        for key in argspec.kwonly_defaults.keys():
            argspec.kwonly_defaults[key] = _Repr(kwparsed[key])

    return argspec


_getpydocspec_re = LazyReCompile(
    r"([a-zA-Z_][a-zA-Z0-9_]*?)\((.*?)\)", re.DOTALL
)


def _getpydocspec(f: Callable) -> ArgSpec | None:
    try:
        argspec = pydoc.getdoc(f)
    except NameError:
        return None

    s = _getpydocspec_re.search(argspec)
    if s is None:
        return None

    if not hasattr_safe(f, "__name__") or s.groups()[0] != f.__name__:
        return None

    args = []
    defaults = []
    varargs = varkwargs = None
    kwonly_args = []
    kwonly_defaults = {}
    for arg in s.group(2).split(","):
        arg = arg.strip()
        if arg.startswith("**"):
            varkwargs = arg[2:]
        elif arg.startswith("*"):
            varargs = arg[1:]
        elif arg == "...":
            # At least print denotes "..." as separator between varargs and kwonly args.
            varargs = ""
        else:
            arg, _, default = arg.partition("=")
            if varargs is not None:
                kwonly_args.append(arg)
                if default:
                    kwonly_defaults[arg] = _Repr(default)
            else:
                args.append(arg)
                if default:
                    defaults.append(_Repr(default))

    return ArgSpec(
        args, varargs, varkwargs, defaults, kwonly_args, kwonly_defaults, None
    )


def getfuncprops(func: str, f: Callable) -> FuncProps | None:
    # Check if it's a real bound method or if it's implicitly calling __init__
    # (i.e. FooClass(...) and not FooClass.__init__(...) -- the former would
    # not take 'self', the latter would:
    try:
        func_name = getattr(f, "__name__", None)
    except:
        # if calling foo.__name__ would result in an error
        func_name = None

    try:
        is_bound_method = (
            (inspect.ismethod(f) and f.__self__ is not None)
            or (func_name == "__init__" and not func.endswith(".__init__"))
            or (func_name == "__new__" and not func.endswith(".__new__"))
        )
    except:
        # if f is a method from a xmlrpclib.Server instance, func_name ==
        # '__init__' throws xmlrpclib.Fault (see #202)
        return None
    try:
        argspec = _get_argspec_from_signature(f)
        try:
            argspec = _fix_default_values(f, argspec)
        except KeyError as ex:
            # Parsing of the source failed. If f has a __signature__, we trust it.
            if not hasattr(f, "__signature__"):
                raise ex
        fprops = FuncProps(func, argspec, is_bound_method)
    except (TypeError, KeyError, ValueError):
        argspec_pydoc = _getpydocspec(f)
        if argspec_pydoc is None:
            return None
        if inspect.ismethoddescriptor(f):
            argspec_pydoc.args.insert(0, "obj")
        fprops = FuncProps(func, argspec_pydoc, is_bound_method)
    return fprops


def is_eval_safe_name(string: str) -> bool:
    return all(
        part.isidentifier() and not keyword.iskeyword(part)
        for part in string.split(".")
    )


def _get_argspec_from_signature(f: Callable) -> ArgSpec:
    """Get callable signature from inspect.signature in argspec format.

    inspect.signature is a Python 3 only function that returns the signature of
    a function.  Its advantage over inspect.getfullargspec is that it returns
    the signature of a decorated function, if the wrapper function itself is
    decorated with functools.wraps.

    """
    args = []
    varargs = None
    varkwargs = None
    defaults = []
    kwonly = []
    kwonly_defaults = {}
    annotations = {}

    # We use signature here instead of getfullargspec as the latter also returns
    # self and cls (for class methods).
    signature = inspect.signature(f)
    for parameter in signature.parameters.values():
        if parameter.annotation is not parameter.empty:
            annotations[parameter.name] = parameter.annotation

        if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            args.append(parameter.name)
            if parameter.default is not parameter.empty:
                defaults.append(parameter.default)
        elif parameter.kind == inspect.Parameter.POSITIONAL_ONLY:
            args.append(parameter.name)
        elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            varargs = parameter.name
        elif parameter.kind == inspect.Parameter.KEYWORD_ONLY:
            kwonly.append(parameter.name)
            kwonly_defaults[parameter.name] = parameter.default
        elif parameter.kind == inspect.Parameter.VAR_KEYWORD:
            varkwargs = parameter.name

    return ArgSpec(
        args,
        varargs,
        varkwargs,
        defaults if defaults else None,
        kwonly,
        kwonly_defaults if kwonly_defaults else None,
        annotations if annotations else None,
    )


_get_encoding_line_re = LazyReCompile(r"^.*coding[:=]\s*([-\w.]+).*$")


def get_encoding(obj) -> str:
    """Try to obtain encoding information of the source of an object."""
    for line in inspect.findsource(obj)[0][:2]:
        m = _get_encoding_line_re.search(line)
        if m:
            return m.group(1)
    return "utf8"


def get_encoding_file(fname: str) -> str:
    """Try to obtain encoding information from a Python source file."""
    with open(fname, encoding="ascii", errors="ignore") as f:
        for _ in range(2):
            line = f.readline()
            match = _get_encoding_line_re.search(line)
            if match:
                return match.group(1)
    return "utf8"


def getattr_safe(obj: Any, name: str) -> Any:
    """Side effect free getattr (calls getattr_static)."""
    result = inspect.getattr_static(obj, name)
    # Slots are a MemberDescriptorType
    if isinstance(result, MemberDescriptorType):
        result = getattr(obj, name)
    # classmethods are safe to access (see #966)
    if isinstance(result, (classmethod, staticmethod)):
        result = result.__get__(obj, obj)
    return result


def hasattr_safe(obj: Any, name: str) -> bool:
    try:
        getattr_safe(obj, name)
        return True
    except AttributeError:
        return False



import gettext
import locale
import os.path
import sys
from typing import Optional, cast, List

from .. import package_dir

translator: gettext.NullTranslations = cast(gettext.NullTranslations, None)


def _(message) -> str:
    return translator.gettext(message)


def ngettext(singular, plural, n):
    return translator.ngettext(singular, plural, n)


def init(
    locale_dir: str | None = None, languages: list[str] | None = None
) -> None:
    try:
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error:
        # This means that the user's environment is broken. Let's just continue
        # with the default C locale.
        sys.stderr.write(
            "Error: Your locale settings are not supported by "
            "the system. Using the fallback 'C' locale instead. "
            "Please fix your locale settings.\n"
        )

    global translator
    if locale_dir is None:
        locale_dir = os.path.join(package_dir, "translations")

    translator = gettext.translation(
        "bpython", locale_dir, languages, fallback=True
    )



import os
import unittest

try:
    import watchdog
    from bpython.curtsiesfrontend.filewatch import ModuleChangedEventHandler

    has_watchdog = True
except ImportError:
    has_watchdog = False

from unittest import mock


@unittest.skipUnless(has_watchdog, "watchdog required")
class TestModuleChangeEventHandler(unittest.TestCase):
    def setUp(self):
        self.module = ModuleChangedEventHandler([], 1)
        self.module.observer = mock.Mock()

    def test_create_module_handler(self):
        self.assertIsInstance(self.module, ModuleChangedEventHandler)

    def test_add_module(self):
        self.module._add_module("something/test.py")
        self.assertIn(
            os.path.abspath("something/test"),
            self.module.dirs[os.path.abspath("something")],
        )

    def test_activate_throws_error_when_already_activated(self):
        self.module.activated = True
        with self.assertRaises(ValueError):
            self.module.activate()



import os
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence

from .. import importcompletion

try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler, FileSystemEvent
except ImportError:

    def ModuleChangedEventHandler(*args):
        return None

else:

    class ModuleChangedEventHandler(FileSystemEventHandler):  # type: ignore [no-redef]
        def __init__(
            self,
            paths: Iterable[str],
            on_change: Callable[[Sequence[str]], None],
        ) -> None:
            self.dirs: dict[str, set[str]] = defaultdict(set)
            self.on_change = on_change
            self.modules_to_add_later: list[str] = []
            self.observer = Observer()
            self.started = False
            self.activated = False
            for path in paths:
                self._add_module(path)

            super().__init__()

        def reset(self) -> None:
            self.dirs.clear()
            self.modules_to_add_later.clear()
            self.observer.unschedule_all()

        def _add_module(self, path: str) -> None:
            """Add a python module to track changes"""
            path = os.path.abspath(path)
            for suff in importcompletion.SUFFIXES:
                if path.endswith(suff):
                    path = path[: -len(suff)]
                    break
            dirname = os.path.dirname(path)
            if dirname not in self.dirs:
                self.observer.schedule(self, dirname, recursive=False)
            self.dirs[dirname].add(path)

        def _add_module_later(self, path: str) -> None:
            self.modules_to_add_later.append(path)

        def track_module(self, path: str) -> None:
            """
            Begins tracking this if activated, or remembers to track later.
            """
            if self.activated:
                self._add_module(path)
            else:
                self._add_module_later(path)

        def activate(self) -> None:
            if self.activated:
                raise ValueError(f"{self!r} is already activated.")
            if not self.started:
                self.started = True
                self.observer.start()
            for dirname in self.dirs:
                self.observer.schedule(self, dirname, recursive=False)
            for module in self.modules_to_add_later:
                self._add_module(module)
            self.modules_to_add_later.clear()
            self.activated = True

        def deactivate(self) -> None:
            if not self.activated:
                raise ValueError(f"{self!r} is not activated.")
            self.observer.unschedule_all()
            self.activated = False

        def on_any_event(self, event: FileSystemEvent) -> None:
            dirpath = os.path.dirname(event.src_path)
            if any(
                event.src_path == f"{path}.py" for path in self.dirs[dirpath]
            ):
                self.on_change((event.src_path,))
